@@ -11,8 +11,7 @@
 #include <sys/types.h>
 #include <glibmm/spawn.h>
 #include "internal/unistd.h"
-
-typedef websocketpp::client<websocketpp::config::asio_client> client;
+#include "CdbPython.hh"
 
 using namespace cadabra;
 
@@ -49,52 +48,26 @@ void ComputeThread::set_master(GUIBase *b, DocumentThread *d)
 void ComputeThread::init()
 	{
 	// Setup the WebSockets client.
-
-	wsclient.init_asio();
-	wsclient.start_perpetual();
 	}
 
 void ComputeThread::try_connect()
 	{
-	using websocketpp::lib::bind;
-
-	// Make the resolver work when there is no network up at all, only localhost.
-	// https://svn.boost.org/trac/boost/ticket/2456
-	// Not sure why this works here, as the compiler claims this statement does
-	// not have any effect.
-	// No longer necessary with websocketpp-0.6.0.
-	//
-	// boost::asio::ip::resolver_query_base::flags(0);
-
-	wsclient.clear_access_channels(websocketpp::log::alevel::all);
-	wsclient.clear_error_channels(websocketpp::log::elevel::all);
-
-	wsclient.set_open_handler(bind(&ComputeThread::on_open,
-	                               this, websocketpp::lib::placeholders::_1));
-	wsclient.set_fail_handler(bind(&ComputeThread::on_fail,
-	                               this, websocketpp::lib::placeholders::_1));
-	wsclient.set_close_handler(bind(&ComputeThread::on_close,
-	                                this, websocketpp::lib::placeholders::_1));
-	wsclient.set_message_handler(bind(&ComputeThread::on_message, this,
-	                                  websocketpp::lib::placeholders::_1,
-	                                  websocketpp::lib::placeholders::_2));
+	wsclient.set_connect_handler(std::bind(&ComputeThread::on_open, this));
+	wsclient.set_fail_handler(std::bind(&ComputeThread::on_fail, this, std::placeholders::_1));
+	wsclient.set_close_handler(std::bind(&ComputeThread::on_close, this));
+	wsclient.set_message_handler(std::bind(&ComputeThread::on_message, this, std::placeholders::_1));
 
 	std::ostringstream uristr;
-	uristr << "ws://" << forced_server_ip_address << ":" << port;
-	websocketpp::lib::error_code ec;
-	connection = wsclient.get_connection(uristr.str(), ec);
-	if (ec) {
-		std::cerr << "cadabra-client: websocket connection error " << ec.message() << std::endl;
-		return;
-		}
+	uristr << "ws://" << (forced_server_ip_address.empty() ? "127.0.0.1" : forced_server_ip_address) << ":" << port;
+	wsclient.connect(uristr.str());
 
-	our_connection_hdl = connection->get_handle();
-	wsclient.connect(connection);
 	// std::cerr << "cadabra-client: connect done" << std::endl;
 	}
 
 void ComputeThread::run()
 	{
+	// This does *not* run on the GUI thread.
+	
 	init();
 	try_spawn_server();
 	try_connect();
@@ -130,11 +103,16 @@ void ComputeThread::terminate()
 		}
 	}
 
+bool ComputeThread::kernel_is_connected() const
+	{
+	return connection_is_open;
+	}
+
 void ComputeThread::all_cells_nonrunning()
 	{
 	for(auto it: running_cells) {
 		std::shared_ptr<ActionBase> rs_action =
-		   std::make_shared<ActionSetRunStatus>(it.second->id(), false);
+		   std::make_shared<ActionSetRunStatus>(it.first, false);
 		docthread->queue_action(rs_action);
 		}
 	if(gui) {
@@ -144,9 +122,9 @@ void ComputeThread::all_cells_nonrunning()
 	running_cells.clear();
 	}
 
-void ComputeThread::on_fail(websocketpp::connection_hdl )
+void ComputeThread::on_fail(const boost::beast::error_code& ec)
 	{
-	std::cerr << "cadabra-client: connection to server on port " << port << " failed" << std::endl;
+	std::cerr << "cadabra-client: connection to server on port " << port << " failed, " << ec.message() << std::endl;
 	connection_is_open=false;
 	all_cells_nonrunning();
 	if(gui && server_pid!=0) {
@@ -182,9 +160,17 @@ void ComputeThread::try_spawn_server()
 
 	std::vector<std::string> argv, envp;
 #if defined(_WIN32) || defined(_WIN64)
-	argv.push_back(cadabra::install_prefix()+"\\bin\\cadabra-server.exe");
+	argv.push_back("cadabra-server.exe");
 #else
-	argv.push_back("cadabra-server");
+	const char *appdir = getenv("APPDIR");
+	if(appdir) {
+		// std::cerr << "This is an AppImage, APPDIR = " << appdir << std::endl;
+		argv.push_back(std::string(appdir)+"/usr/bin/cadabra-server");
+		}
+	else {
+		// std::cerr << "Not an AppImage." << std::endl;
+		argv.push_back("cadabra-server");
+		}
 #endif
 	Glib::Pid pid;
 	std::string wd("");
@@ -192,15 +178,38 @@ void ComputeThread::try_spawn_server()
 	// See https://bugs.launchpad.net/inkscape/+bug/1662531 for things related to
 	// the 'envp' argument in the call below.
 	try {
+#ifdef _WIN32
+		Glib::SpawnFlags flags = Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_SEARCH_PATH | Glib::SpawnFlags::SPAWN_STDERR_TO_DEV_NULL;
+#else
+  #if GLIBMM_MAJOR_VERSION > 2 || (GLIBMM_MAJOR_VERSION == 2 && GLIBMM_MINOR_VERSION >= 68)
+		Glib::SpawnFlags flags = Glib::SpawnFlags::DEFAULT | Glib::SpawnFlags::SEARCH_PATH,
+  #else
+		Glib::SpawnFlags flags = Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_SEARCH_PATH;
+  #endif
+#endif
+		
 		Glib::spawn_async_with_pipes(wd, argv, /* envp, WITH envp, Fedora 27 fails to start python properly */
-		                             Glib::SPAWN_DEFAULT|Glib::SPAWN_SEARCH_PATH,
-		                             sigc::slot<void>(),
+											  flags,
+		                             sigc::slot<void()>(),
 		                             &pid,
 		                             0,
 		                             &server_stdout,
 		                             0); // We need to see stderr on the console
 		//										  &server_stderr);
 
+//#if GLIBMM_MAJOR_VERSION > 2 || (GLIBMM_MAJOR_VERSION == 2 && GLIBMM_MINOR_VERSION >= 68)
+//											  Glib::SpawnFlags::DEFAULT | Glib::SpawnFlags::SEARCH_PATH,
+//		                             sigc::slot<void()>(),
+//#else
+//		                             flags,
+//		                             sigc::slot<void>([](){ FreeConsole(); }),
+//#endif
+//		                             &pid,
+//		                             0,
+//		                             &server_stdout,
+//		                             0); // We need to see stderr on the console
+//		//										  &server_stderr);
+
 		char buffer[100];
 		FILE *f = fdopen(server_stdout, "r");
 		if(fscanf(f, "%99s", buffer)!=1) {
@@ -219,12 +228,14 @@ void ComputeThread::try_spawn_server()
 		}
 	}
 
-void ComputeThread::on_open(websocketpp::connection_hdl )
+void ComputeThread::on_open()
 	{
 	connection_is_open=true;
 	restarting_kernel=false;
-	if(gui)
+	if(gui) {
 		gui->on_connect();
+		gui->on_kernel_runstatus(false);
+		}
 
 	//	// now it is safe to use the connection
 	//	std::string msg;
@@ -246,14 +257,19 @@ void ComputeThread::on_open(websocketpp::connection_hdl )
 	//	c->send(hdl, msg, websocketpp::frame::opcode::text);
 	}
 
-void ComputeThread::on_close(websocketpp::connection_hdl )
+void ComputeThread::on_close()
 	{
 	// std::cerr << "cadabra-client: connection closed" << std::endl;
 	connection_is_open=false;
 	all_cells_nonrunning();
 	if(gui) {
-		if(restarting_kernel) gui->on_disconnect("restarting kernel");
-		else                  gui->on_disconnect("not connected");
+		if(restarting_kernel) {
+			gui->on_disconnect("restarting kernel");
+			gui->on_kernel_runstatus(true);
+			}
+		else {
+			gui->on_disconnect("not connected");
+			}
 		}
 
 	sleep(1); // do not cause a torrent...
@@ -262,30 +278,45 @@ void ComputeThread::on_close(websocketpp::connection_hdl )
 
 void ComputeThread::cell_finished_running(DataCell::id_t id)
 	{
-	auto it=running_cells.find(id);
-	if(it==running_cells.end()) {
-		throw std::logic_error("Cannot find cell with id = "+std::to_string(id.id));
+	if(id.id==0) {
+		// This was code without a cell representation (run in
+		// response to e.g. a slider update). Ignore.
 		}
+	else {
+		auto it=running_cells.find(id);
+		if(it==running_cells.end()) {
+			throw std::logic_error("Cannot find cell with id = "+std::to_string(id.id));
+			}
+		
+		if(it->second==1) {
+			// Mark this cell as no longer running.
+			std::shared_ptr<ActionBase> rs_action =
+				std::make_shared<ActionSetRunStatus>(id, false);
+			docthread->queue_action(rs_action);
 
-	//	DTree::iterator ret = (*it).second;
-	running_cells.erase(it);
+			running_cells.erase(it);
+			}
+		else
+			it->second -= 1;
+		}
 	}
 
-void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
+void ComputeThread::on_message(const std::string& msg)
 	{
-	client::connection_ptr con = wsclient.get_con_from_hdl(hdl);
-	// std::cerr << "payload: " << msg->get_payload() << std::endl;
-
 	// Parse the JSON message.
 	nlohmann::json root;
 
 	try {
-		root=nlohmann::json::parse(msg->get_payload());
+		root=nlohmann::json::parse(msg);
 		}
 	catch(nlohmann::json::exception& e) {
 		std::cerr << "cadabra-client: cannot parse message." << std::endl;
 		return;
 		}
+	if(getenv("CADABRA_SHOW_RECEIVED")) {
+		std::cerr << "RECV: " << root.dump(3) << std::endl;
+		}
+	
 	if(root.count("header")==0) {
 		std::cerr << "cadabra-client: received message without 'header'." << std::endl;
 		return;
@@ -311,6 +342,8 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 	else
 		cell_id.created_by_client=false;
 	// std::cerr << "received cell with id " << cell_id.id << std::endl;
+
+	// Determine if this refers to a special cell in the interactive console.
 	if (interactive_cells.find(parent_id.id) != interactive_cells.end()) {
 		interactive_cells.insert(cell_id.id);
 		docthread->on_interactive_output(root);
@@ -349,9 +382,18 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 			bool finished = header["last_in_sequence"].get<bool>();
 
 			if (finished) {
-				std::shared_ptr<ActionBase> rs_action =
-				   std::make_shared<ActionSetRunStatus>(parent_id, false);
-				docthread->queue_action(rs_action);
+				if(parent_id.id!=0) {
+					// If this cell references variables, store them in the DataCell.
+					if(content.count("variables")>0) {
+						std::set<std::string> vars;
+						for(const auto& var: content["variables"])
+							vars.insert(var);
+						
+						std::shared_ptr<ActionBase> action =
+							std::make_shared<ActionSetVariableList>(parent_id, vars);
+						docthread->queue_action(action);
+						}
+					}
 				cell_finished_running(parent_id);
 				}
 
@@ -395,12 +437,12 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
-				else if (msg_type == "error") {
-					std::string error = "{\\color{red}{\\begin{verbatim}" + content["output"].get<std::string>()
-					                    + "\\end{verbatim}}}";
-					if (msg_type == "fault") {
-						error = "{\\color{red}{Kernel fault}}\\begin{small}" + error + "\\end{small}";
-						}
+				else if (msg_type == "error" || msg_type=="fault") {
+					std::string error = "\\begin{verbatim}" + content["output"].get<std::string>()
+					                    + "\\end{verbatim}";
+					//if (msg_type == "fault") {
+					//	error = "Kernel fault\\begin{small}" + error + "\\end{small}";
+					//	}
 
 					// Stick an AddCell action onto the stack. We instruct the
 					// action to add this result output cell as a child of the
@@ -426,9 +468,21 @@ void ComputeThread::on_message(websocketpp::connection_hdl hdl, message_ptr msg)
 					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
 					docthread->queue_action(action);
 					}
+				else if (msg_type == "image_svg") {
+					DataCell result(cell_id, DataCell::CellType::image_svg, content["output"].get<std::string>());
+					std::shared_ptr<ActionBase> action =
+					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
+					docthread->queue_action(action);
+					}
+				else if (msg_type == "slider") {
+					DataCell result(cell_id, DataCell::CellType::slider, content["output"].get<std::string>());
+					std::shared_ptr<ActionBase> action =
+					   std::make_shared<ActionAddCell>(result, parent_id, ActionAddCell::Position::child);
+					docthread->queue_action(action);
+					}
 				else {
 					std::cerr << "cadabra-client: received cell we did not expect: "
-					          << msg_type << std::endl;
+					          << msg_type << ": " << content << std::endl;
 					}
 				}
 			}
@@ -470,11 +524,14 @@ void ComputeThread::execute_interactive(uint64_t id, const std::string& code)
 
 	std::ostringstream oss;
 	oss << req << std::endl;
-	wsclient.send(our_connection_hdl, oss.str(), websocketpp::frame::opcode::text);
+	if(getenv("CADABRA_SHOW_SENT")) {
+		std::cerr << "SEND: " << req.dump(3) << std::endl;
+		}
+	wsclient.send(oss.str());
 	interactive_cells.insert(id);
 	}
 
-void ComputeThread::execute_cell(DTree::iterator it)
+void ComputeThread::execute_cell(DTree::iterator it, std::string no_assign, std::vector<uint64_t> output_cell_ids)
 	{
 	// This absolutely has to be run on the main GUI thread.
 	assert(gui_thread_id==std::this_thread::get_id());
@@ -506,7 +563,11 @@ void ComputeThread::execute_cell(DTree::iterator it)
 	// For a code cell, construct a server request message and then
 	// send the cell to the server.
 	if(it->cell_type==DataCell::CellType::python) {
-		running_cells[dc.id()]=it;
+		auto rit=running_cells.find(dc.id());
+		if(rit==running_cells.end())
+			running_cells[dc.id()]=1;
+		else
+			rit->second += 1;
 
 		// Schedule an action to update the running status of this cell.
 		std::shared_ptr<ActionBase> rs_action =
@@ -521,15 +582,20 @@ void ComputeThread::execute_cell(DTree::iterator it)
 		else
 			header["cell_origin"]="server";
 		header["msg_type"]="execute_request";
+		header["output_cell_ids"]=output_cell_ids;
 		req["auth_token"]=authentication_token;
 		req["header"]=header;
+		content["remove_variable_assignments"]=no_assign;
 		content["code"]=dc.textbuf;
 		req["content"]=content;
 
 		gui->on_kernel_runstatus(true);
 		std::ostringstream str;
 		str << req << std::endl;
-		wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+		if(getenv("CADABRA_SHOW_SENT")) {
+			std::cerr << "SEND: " << req.dump(3) << std::endl;
+			}
+		wsclient.send(str.str());
 		// NOTE: we can get a return message in on_message at any point after this,
 		// it will come in on a different thread!
 		}
@@ -547,6 +613,27 @@ void ComputeThread::execute_cell(DTree::iterator it)
 		}
 	}
 
+void ComputeThread::update_variable_on_server(std::string variable, double value)
+	{
+	nlohmann::json req, header, content;
+	header["uuid"]="none";
+	header["cell_id"]=0;
+	header["cell_origin"]="client";
+	header["msg_type"]="execute_request";
+	req["auth_token"]=authentication_token;
+	req["header"]=header;
+	content["code"]=variable + "=" + std::to_string(value);
+	req["content"]=content;
+	
+	std::ostringstream str;
+	str << req << std::endl;
+	if(getenv("CADABRA_SHOW_SENT")) {
+		std::cerr << "SEND: " << req.dump(3) << std::endl;
+		}
+	wsclient.send(str.str());
+	}
+
+
 int ComputeThread::number_of_cells_executing() const
 	{
 	return running_cells.size();
@@ -569,7 +656,11 @@ void ComputeThread::stop()
 	//	std::cerr << str.str() << std::endl;
 
 	server_pid=0;
-	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+	if(getenv("CADABRA_SHOW_SENT")) {
+		std::cerr << "SEND: " << req.dump(3) << std::endl;
+		}
+	wsclient.send(str.str());
+	all_cells_nonrunning();	
 	}
 
 void ComputeThread::restart_kernel()
@@ -597,7 +688,10 @@ void ComputeThread::restart_kernel()
 
 	//	std::cerr << str.str() << std::endl;
 
-	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+	if(getenv("CADABRA_SHOW_SENT")) {
+		std::cerr << "SEND: " << req.dump(3) << std::endl;
+		}
+	wsclient.send(str.str());
 	docthread->on_interactive_output(req);
 	}
 
@@ -640,7 +734,10 @@ bool ComputeThread::complete(DTree::iterator it, int pos, int alternative)
 	// std::cerr << str.str() << std::endl;
 
 	server_pid=0;
-	wsclient.send(our_connection_hdl, str.str(), websocketpp::frame::opcode::text);
+	if(getenv("CADABRA_SHOW_SENT")) {
+		std::cerr << "SENT: " << req.dump(3) << std::endl;
+		}
+	wsclient.send(str.str());
 
 	return true;
 	}
