@@ -2,6 +2,8 @@
 #include "Actions.hh"
 #include "DocumentThread.hh"
 #include "GUIBase.hh"
+#include "ComputeThread.hh"
+#include "CdbPython.hh"
 
 #include <typeinfo>
 #include <iostream>
@@ -22,10 +24,16 @@
 using namespace cadabra;
 
 DocumentThread::DocumentThread(GUIBase* g)
-	: gui(g), compute(0), disable_stacks(false)
+	: gui(g)
+	, compute(0)
+	, current_cell(doc.end())
+	, follow_mode(false)
+	, follow_cell(doc.end())
+	, follow_last_cell(doc.end())
+	, disable_stacks(false)
 	{
 	// Setup logging.
-	std::string version=std::string(CADABRA_VERSION_MAJOR)+"."+CADABRA_VERSION_MINOR+"."+CADABRA_VERSION_PATCH;
+	std::string version=std::string(CADABRA_VERSION_SEM);
 #ifndef EMSCRIPTEN
 	snoop::log.init("Cadabra", version, "log.cadabra.science");
 	snoop::log.set_sync_immediately(true);
@@ -75,7 +83,7 @@ void DocumentThread::new_document()
 void DocumentThread::load_from_string(const std::string& json)
 	{
 	std::lock_guard<std::mutex> guard(stack_mutex);
-	pending_actions=std::queue<std::shared_ptr<ActionBase> >(); // clear queue
+	pending_actions=std::queue<std::shared_ptr<ActionBase> >();
 	doc.clear();
 	JSON_deserialise(json, doc);
 	gui->remove_all_cells();
@@ -97,12 +105,34 @@ void DocumentThread::undo()
 
 	redo_stack.push(ua);
 	undo_stack.pop();
+	// std::cerr << "DocumentThread::undo: undo_stack.size() == " << undo_stack.size() << std::endl;
 	ua->revert(*this, *gui);
 	disable_stacks=false;
 
 	stack_mutex.unlock();
 	}
 
+void DocumentThread::redo()
+	{
+	stack_mutex.lock();
+	if(redo_stack.size()==0) {
+		//std::cerr << "no entries left on the stack" << std::endl;
+		stack_mutex.unlock();
+		return;
+		}
+
+	disable_stacks=true;
+	auto ua = redo_stack.top();
+	//std::cerr << "Undo action " << typeid(*ua).name() << std::endl;
+
+	undo_stack.push(ua);
+	redo_stack.pop();
+	ua->execute(*this, *gui);
+	disable_stacks=false;
+
+	stack_mutex.unlock();
+	}
+
 void DocumentThread::build_visual_representation()
 	{
 	// Because the add_cell method figures out by itself where to generate the VisualCell,
@@ -110,7 +140,7 @@ void DocumentThread::build_visual_representation()
 
 	DTree::iterator doc_it=doc.begin();
 	while(doc_it!=doc.end()) {
-		//		std::cout << doc_it->textbuf << std::endl;
+		// std::cout << "ADDING:" << doc_it->textbuf << std::endl;
 		gui->add_cell(doc, doc_it, false);
 		++doc_it;
 		}
@@ -168,14 +198,105 @@ void DocumentThread::queue_action(std::shared_ptr<ActionBase> ab)
 	pending_actions.push(ab);
 	}
 
+void DocumentThread::run_all_cells()
+	{
+	follow_mode=true;
+	DTree::sibling_iterator sib=doc.begin(doc.begin());
+	while(sib!=doc.end(doc.begin())) {
+		if(sib->cell_type==DataCell::CellType::python) {
+			run_cell(DTree::iterator(sib), false);
+			follow_last_cell=DTree::iterator(sib);
+			}			
+		++sib;
+		}
+	}
+
+void DocumentThread::run_cell(DTree::iterator it, bool shift_pressed)
+	{
+	// First ensure that this cell is not already running, otherwise all hell
+	// will break loose when we try to double-remove the existing output cell etc.
+
+	if(it->running)
+		return;
+
+	// Ensure this cell is not empty either.
+
+	if(it->textbuf.size()==0)
+		return;
+
+	// Remove child nodes, if any.
+	// FIXME: Does it make more sense to do this only after the
+	// execution result comes back from the server?
+
+	DTree::sibling_iterator sib=doc.begin(it);
+	gui->dim_output_cells(it);
+	while(sib!=doc.end(it)) {
+		// std::cout << "cadabra-client: scheduling output cell for removal: " << sib->id().id << std::endl;
+		std::shared_ptr<ActionBase> action = std::make_shared<ActionRemoveCell>(sib->id());
+		queue_action(action);
+		++sib;
+		}
+
+	// Since the user has initiated this cell execution, we can
+	// turn on cell follow mode.
+	follow_cell=it;
+	follow_mode=true;
+
+	// Execute the cell.
+	// std::cerr << "Executing cell " << it->id().id << std::endl;
+
+	// If this is a LaTeX input cell, and auto-close is turned on, close
+	// the input cell. Make sure to also feed that into the document
+	// itself!
+	if(it->cell_type==DataCell::CellType::latex) 
+		if(prefs.auto_close_latex) 
+			gui->hide_visual_cells(it);
+	
+	// Execute the cell. Make sure this comes after the hiding logic above.
+	compute->execute_cell(it);
+	}
+
+
+void DocumentThread::run_cells_referencing_variable(std::string variable, double value)
+	{
+	// First update the variable itself.
+	compute->update_variable_on_server(variable, value);
+
+	// Re-run all cells referencing this variable.
+	follow_mode=false;
+	DTree::iterator it = doc.begin();
+	while(it!=doc.end()) {
+		if(it->cell_type==DataCell::CellType::python) {
+			if(it->variables_referenced.count(variable)==1) {
+				if(it->textbuf.find("slider(")==std::string::npos) {
+					// We have found a cell which depends on the variable.
+					// Collect the cell_id's of the current output cells,
+					// so that we can re-use these.
+					std::vector<uint64_t> output_cell_ids;
+					DTree::iterator sib = doc.begin(it);
+					while(sib != doc.end(it)) {
+						output_cell_ids.push_back(sib->id().id);
+						++sib;
+						}
+					
+					// Now execute.
+					compute->execute_cell(it, variable, output_cell_ids);
+					}
+				}
+			}
+		++it;
+		}
+	}
 
 void DocumentThread::process_action_queue()
 	{
 	// FIXME: we certainly do not want any two threads to run this at the same time,
 	// but that is not guaranteed. Actions should always be run on the GUI thread.
 	// This absolutely has to be run on the main GUI thread.
-	//	assert(main_thread_id==std::this_thread::get_id());
 
+	if(main_thread_id != std::this_thread::get_id())
+		std::cerr << "INTERNAL ERROR: DocumentThread::process_action_queue not running on main thread."
+					 << std::endl;
 
 	stack_mutex.lock();
 	while(pending_actions.size()>0) {
@@ -198,7 +319,10 @@ void DocumentThread::process_action_queue()
 		stack_mutex.lock();
 		if(ab->undoable())
 			undo_stack.push(ab);
-		pending_actions.pop();
+		if(ab->callback != nullptr)
+			ab->callback();
+		if(pending_actions.size()>0) // some actions clear the queue
+			pending_actions.pop();
 		}
 	stack_mutex.unlock();
 	}
@@ -256,6 +380,17 @@ DocumentThread::Prefs::Prefs(bool use_defaults)
 	python_path        = data.value("python_path", "");
 	move_into_new_cell = data.value("move_into_new_cell", false);
 	tab_completion     = data.value("tab_completion", true);
+	microtex           = data.value("microtex", true);
+	auto_close_latex   = data.value("auto_close_latex", true);
+
+	// Force microtex when this is an AppImage.
+	const char *appdir = getenv("APPDIR");
+	if(appdir)
+		microtex=true;
+	// Force microtex when we are on Windows.
+#if(_WIN32)
+	microtex = true;
+#endif
 
 	if(git_path=="")
 		git_path="/usr/bin/git";
@@ -297,6 +432,8 @@ void DocumentThread::Prefs::save()
 		data["python_path"] = python_path;
 		data["move_into_new_cell"] = move_into_new_cell;
 		data["tab_completion"] = tab_completion;
+		data["microtex"] = microtex;
+		data["auto_close_latex"] = auto_close_latex;
 		for (const auto& lang : colours) {
 			for (const auto& kw : lang.second)
 				data["colours"][lang.first][kw.first] = kw.second;
