@@ -37,6 +37,10 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include "properties/Symbol.hh"
 #include "properties/Trace.hh"
 #include "properties/DependsBase.hh"
+#include "properties/NonCommuting.hh"
+#include "properties/AntiCommuting.hh"
+#include "properties/SelfNonCommuting.hh"
+#include "properties/SelfAntiCommuting.hh"
 
 #include <sstream>
 
@@ -337,7 +341,12 @@ void Algorithm::propagate_zeroes(post_order_iterator& it, const iterator& topnod
 		// Removing happens in the next step.
 		}
 	else if(*walk->name=="\\pow") {
-		if(tr.index(it)==0) { // the argument
+		if(tr.index(it)==0) { // the argument is zero
+			sibling_iterator arg=it;
+			++arg;
+			if(*arg->multiplier<0) 
+				throw RuntimeException("Division by zero encountered.");
+			
 			walk->multiplier=rat_set.insert(0).first;
 			it=walk;
 			propagate_zeroes(it, topnode);
@@ -871,6 +880,23 @@ bool Algorithm::is_nonprod_factor_in_prod(iterator it)
 	return false;
 	}
 
+bool Algorithm::is_noncommuting(const Properties& properties, iterator it)
+	{
+	auto nc = properties.get<NonCommuting>(it);
+	if(nc) return true;
+
+	auto ac = properties.get<AntiCommuting>(it);
+	if(ac) return true;
+	
+	auto snc = properties.get<SelfNonCommuting>(it);
+	if(snc) return true;
+	
+	auto sac = properties.get<SelfAntiCommuting>(it);
+	if(sac) return true;
+
+	return false;
+	}
+
 bool Algorithm::prod_wrap_single_term(iterator& it)
 	{
 	if(is_single_term(it)) {
@@ -1055,7 +1081,7 @@ bool Algorithm::locate_object_set(const Ex& objs,
 		if((*aim->name)=="\\comma") {
 			// Objects can themselves be lists of other objects (for instance
 			// when we want to symmetrise in index pairs).
-			if(locate_object_set(aim, st, nd, store)==false)
+			if(locate_object_set(Ex(aim), st, nd, store)==false)
 				return false;
 			}
 		else {
