@@ -17,15 +17,8 @@
 #include "properties/Integer.hh"
 #include "properties/SortOrder.hh"
 
-// #define DEBUG 1
-
-static bool debug_stop = false;
-
-#ifdef DEBUG
-#define DEBUGLN(ln) if(!debug_stop) { ln; }
-#else
-#define DEBUGLN(ln)
-#endif
+// #define DEBUG "Compare.cc"
+#include "Debug.hh"
 
 namespace cadabra {
 
@@ -158,6 +151,11 @@ namespace cadabra {
 		return subtree_exact_less(properties, one.begin(), two.begin(), mod_prel, checksets, compare_multiplier, literal_wildcards);
 		}
 
+	bool tree_exact_less(const Properties* properties, Ex::iterator one, Ex::iterator two, int mod_prel, bool checksets, int compare_multiplier, bool literal_wildcards)
+		{
+		return subtree_exact_less(properties, one, two, mod_prel, checksets, compare_multiplier, literal_wildcards);
+		}
+
 	bool tree_exact_equal(const Properties* properties, const Ex& one, const Ex& two, int mod_prel, bool checksets, int compare_multiplier, bool literal_wildcards)
 		{
 		return subtree_exact_equal(properties, one.begin(), two.begin(), mod_prel, checksets, compare_multiplier, literal_wildcards);
@@ -233,6 +231,11 @@ namespace cadabra {
 		return tree_exact_less(properties, one, two);
 		}
 
+	bool tree_exact_less_obj::operator()(Ex::iterator one, Ex::iterator two) const
+		{
+		return tree_exact_less(properties, one, two);
+		}
+
 	tree_exact_less_no_wildcards_obj::tree_exact_less_no_wildcards_obj()
 		{
 		properties=0;
@@ -454,7 +457,7 @@ namespace cadabra {
 #endif
 
 		DEBUGLN( std::cerr << tab() << "result = "; );
-		if(!debug_stop) {
+		DEBUGLN( 
 			switch(r) {
 				case match_t::node_match:
 					std::cerr << "node_match";
@@ -481,8 +484,7 @@ namespace cadabra {
 					std::cerr << "no_match_greater";
 					break;
 				}
-			std::cerr << std::endl;
-			}
+			std::cerr << std::endl; );
 		
 		--offset;
 		return r;
@@ -593,7 +595,7 @@ namespace cadabra {
 			// triggering a rule for an upper index, we simply store both rules (see
 			// below) so that searching for rules can remain simple.
 
-			replacement_map_t::iterator loc=replacement_map.find(one);
+			replacement_map_t::iterator loc=replacement_map.find(Ex(one));
 
 			bool tested_full=true;
 
@@ -651,15 +653,15 @@ namespace cadabra {
 					const Indices *t2=0;
 					if(use_props==useprops_t::always) {
 						DEBUGLN( std::cerr << tab() << "is " << *one->name << " an index?" << std::endl; );
-						debug_stop=true;
+						DEBUGSTOP(true);
 						t1=properties.get<Indices>(one, false);
-						debug_stop=false;						
+						DEBUGSTOP(false);
 						DEBUGLN( std::cerr << tab() << "found index property for one: " << t1 << std::endl; );
 						if(two->is_rational()==false) {
 							DEBUGLN( std::cerr << tab() << "is " << *two->name << " an index?" << std::endl; );
-							debug_stop=true;
+							DEBUGSTOP(true);
 							t2=properties.get<Indices>(two, false);
-							debug_stop=false;
+							DEBUGSTOP(false);
 							DEBUGLN( std::cerr << tab() << "found index property for two: " << t2 << std::endl; );
 							// It is still possible that t2 is a Coordinate and
 							// t1 an Index which can take the value of the
@@ -672,12 +674,14 @@ namespace cadabra {
 							// evaluate tries to determine if a rule for X^{t}
 							// applies to an expression X^{m}).
 							if(t1!=0 && t2!=t1) {
-								auto ivals = std::find_if(t1->values.begin(), t1->values.end(),
-								[&](const Ex& a) {
-									if(subtree_compare(&properties, a.begin(), two, 0)==0) return true;
-									else return false;
-								});
-								if(ivals!=t1->values.end()) {
+								const auto& values = t1->values(properties, one);
+								auto ivals = std::find_if(values.begin(), values.end(),
+																  [&](const Ex& a) {
+																  if(subtree_compare(&properties, a.begin(), two, 0)==0)
+																	  return true;
+																  else return false;
+																  });
+								if(ivals!=values.end()) {
 									t2=t1;
 									two_is_value=true;
 									}
@@ -733,9 +737,15 @@ namespace cadabra {
 				// that e.g a found _{z} also leads to a replacement rule for (z), or
 				// if we want that a found _{z} also leads to a replacement for ^{z},
 				// this needs to be added to the replacement map explicitly.
+				
+				DEBUGLN( std::cerr << "adding " << one << " -> " << two << " to replacement map " << std::endl; );
 
-					DEBUGLN( std::cerr << "adding " << one << " -> " << two << " to replacement map " << std::endl; );
-				replacement_map[one]=two;
+				// We absolutely need to keep the multiplier here. If we don't, then it
+				// becomes very messy to restore multipliers of child nodes in the pattern,
+				// e.g A?_{m n} matching Q_{1 -1}. Do not set the multiplier to one!
+				Ex tmp(two);
+            // cadabra::one(tmp.begin()->multiplier);
+				replacement_map[Ex(one)]=tmp;
 
 				// if this is an index, also store the pattern with the parent_rel flipped
 
@@ -799,6 +809,7 @@ namespace cadabra {
 			else return report(match_t::node_match);
 			}
 		else if(objectpattern) {
+			DEBUGLN( std::cerr << "object pattern " << one << std::endl; );
 			// Even for object patterns, if we do not ignore the parent rel, we need to test them.
 			if(!ignore_parent_rel)
 				if( one->fl.parent_rel != two->fl.parent_rel )
@@ -809,7 +820,11 @@ namespace cadabra {
 			if(loc!=subtree_replacement_map.end()) {
 				return report(equal_subtree((*loc).second,two,use_props));
 				}
-			else subtree_replacement_map[one->name]=two;
+			else {
+				DEBUGLN( std::cerr << "storing subtree replacement " << one
+							<< " -> " << two << std::endl; );
+				subtree_replacement_map[one->name]=two;
+				}
 
 			return report(match_t::subtree_match);
 			}
@@ -838,12 +853,13 @@ namespace cadabra {
 				// Look through values attribute of Indices object to see if the 'two' index
 				// can take the 'one' value.
 
-				auto ivals = std::find_if(t2->values.begin(), t2->values.end(),
-				[&](const Ex& a) {
-					if(subtree_compare(&properties, a.begin(), one, 0)==0) return true;
-					else return false;
-					});
-				if(ivals!=t2->values.end()) {
+				const auto& values = t2->values(properties, two);
+				auto ivals = std::find_if(values.begin(), values.end(),
+												  [&](const Ex& a) {
+												  if(subtree_compare(&properties, a.begin(), one, 0)==0) return true;
+												  else return false;
+												  });
+				if(ivals!=values.end()) {
 					// Verify that the 'two' index has not already been matched to a value
 					// different from 'one'.
 					Ex t1(two), t2(two), o1(one), o2(one);
@@ -860,7 +876,7 @@ namespace cadabra {
 						return report(match_t::no_match_less);
 						}
 
-					index_value_map[two]=one;
+					index_value_map[Ex(two)]=Ex(one);
 					return report(match_t::node_match);
 					}
 				else {
@@ -1673,7 +1689,7 @@ namespace cadabra {
 					++it;
 					}
 				//			txtout << countpairs << " pairs" << std::endl;
-				if(countpairs!=*(cond.begin()->multiplier))
+				if(countpairs!=cond.begin()->multiplier->get_rational())
 					return false;
 				}
 			else if(*cond->name=="\\regex") {
