@@ -28,6 +28,16 @@ static std::string trim(const std::string& s)
 	return std::string(s, b, e - b + 1);
 	}
 
+DTree::DTree()
+	: tree<DataCell>(), hide_input_cells(false)
+	{
+	}
+
+DTree::DTree(const iterator& it)
+	: tree<DataCell>(it), hide_input_cells(false)
+	{
+	}
+
 bool DataCell::id_t::operator<(const DataCell::id_t& other) const
 	{
 	if(created_by_client != other.created_by_client) return created_by_client;
@@ -91,6 +101,7 @@ std::string cadabra::export_as_HTML(const DTree& doc, bool for_embedding, bool s
 
 std::string cadabra::latex_to_html(const std::string& str)
 	{
+	// std::cerr << "converting " << str << std::endl;
 	std::regex section(R"(\\section\*\{([^\}]*)\})");
 	std::regex author(R"(\\author\{([^\}]*)\})");
 	std::regex email(R"(\\email\{([^\}]*)\})");
@@ -126,10 +137,15 @@ std::string cadabra::latex_to_html(const std::string& str)
 	std::regex ftableau(R"(\\ftableau\{(\{[^\}]*\}[,]?)*\})");
 	std::regex begin_tabular(R"(\\begin\{tabular\}\{[^\}]*\})");
 	std::regex end_tabular(R"(\\end\{tabular\})");
+	std::regex hash_in_math(R"((\\begin\{dmath\*\}.*)\\#(\\end\{dmath\*\}))");
+	std::regex verbatim_symbol(R"(\{\\texttt\{\\backslash\{\}([^\}]*)\}\})");
+//	std::regex verbatim_symbol(R"(\\texttt\{([^\}]*)\})");
 	std::string res;
 
 	try {
-		res = std::regex_replace(str, begin_dmath, R"(\(\displaystyle)");
+		res = std::regex_replace(str, hash_in_math, "$1#$2");
+      res = std::regex_replace(res, verbatim_symbol, "$\\setminus\\texttt{$1}$");
+		res = std::regex_replace(res, begin_dmath, R"(\(\displaystyle)");
 		res = std::regex_replace(res, discretionary, " ");
 		res = std::regex_replace(res, end_dmath, R"(\))");
 		res = std::regex_replace(res, tilde, " ");
@@ -170,6 +186,7 @@ std::string cadabra::latex_to_html(const std::string& str)
 		std::cerr << "regex error on " << str << std::endl;
 		}
 
+	// std::cerr << "to " << res << std::endl;
 	return res;
 	}
 
@@ -179,6 +196,11 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
                            std::string title)
 	{
 	bool strip_this=false;
+
+	if(doc.hide_input_cells)
+		if(it->cell_type==DataCell::CellType::python || it->cell_type==DataCell::CellType::latex)
+			strip_this=true;
+
 	switch(it->cell_type) {
 		case DataCell::CellType::document:
 			if(!for_embedding) {
@@ -216,8 +238,14 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 		case DataCell::CellType::image_png:
 			str << "<div class='image_png'><img src='data:image/png;base64,";
 			break;
+		case DataCell::CellType::image_svg:
+			str << "<div class='image_svg'><img src='data:image/svg+xml;base64,";
+			break;
+		case DataCell::CellType::slider:
+			break;
 		case DataCell::CellType::input_form:
-			str << "<div class='input_form'>";
+			// str << "<div class='input_form'>";
+			strip_this=true;
 			break;
 		}
 
@@ -226,14 +254,23 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 			if(it->textbuf.size()>0) {
 				if(it->cell_type==DataCell::CellType::image_png)
 					str << it->textbuf;
+				else if(it->cell_type==DataCell::CellType::image_svg)
+					str << it->textbuf;
+				else if(it->cell_type==DataCell::CellType::slider) {
+					}
 				else if(it->cell_type!=DataCell::CellType::document && it->cell_type!=DataCell::CellType::latex) {
 					std::string out;
 					if(it->cell_type==DataCell::CellType::python)
 						out=it->textbuf;
 					else
 						out=latex_to_html(it->textbuf);
-					if(out.size()>0)
+					if(out.size()>0) {
+						if(it->cell_type==DataCell::CellType::python) {
+							out = std::regex_replace(out, std::regex("<"), "&lt;");
+							out = std::regex_replace(out, std::regex(">"), "&gt;");
+							}
 						str << "<div class=\"source donthyphenate\">"+out+"</div>";
+						}
 					}
 				}
 			}
@@ -284,8 +321,14 @@ void cadabra::HTML_recurse(const DTree& doc, DTree::iterator it, std::ostringstr
 		case DataCell::CellType::image_png:
 			str << "' /></div>\n";
 			break;
+		case DataCell::CellType::image_svg:
+			str << "' /></div>\n";
+			break;
+		case DataCell::CellType::slider:
+			break;
 		case DataCell::CellType::input_form:
-			str << "</div>\n";
+			// str << "</div>\n";
+			break;
 		}
 	}
 
@@ -303,10 +346,14 @@ std::string cadabra::JSON_serialise(const DTree& doc)
 void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, nlohmann::json& json)
 	{
 	switch(it->cell_type) {
-		case DataCell::CellType::document:
+		case DataCell::CellType::document: {
 			json["description"]="Cadabra JSON notebook format";
 			json["version"]=1.0;
+			nlohmann::json flags;
+			flags["hide_input_cells"] = doc.hide_input_cells;
+			json["flags"]=flags;
 			break;
+			}
 		case DataCell::CellType::python:
 			json["cell_type"]="input";
 			break;
@@ -328,6 +375,12 @@ void cadabra::JSON_recurse(const DTree& doc, DTree::iterator it, nlohmann::json&
 		case DataCell::CellType::image_png:
 			json["cell_type"]="image_png";
 			break;
+		case DataCell::CellType::image_svg:
+			json["cell_type"]="image_svg";
+			break;
+		case DataCell::CellType::slider:
+			json["cell_type"]="slider";
+			break;
 		case DataCell::CellType::input_form:
 			json["cell_type"]="input_form";
 			break;
@@ -378,13 +431,19 @@ void cadabra::JSON_deserialise(const std::string& cj, DTree& doc)
 		root=nlohmann::json::parse(cj);
 		}
 	catch(nlohmann::json::exception& e) {
-		std::cerr << "cannot parse json file" << std::endl;
+		std::cerr << "Cannot parse json file." << std::endl;
 		return;
 		}
 
 	// Setup main document.
 	DataCell::id_t id;
-	id.id=root.value("cell_id", generate_uuid<uint64_t>());
+	try {
+		id.id=root.value("cell_id", generate_uuid<uint64_t>());
+		}
+	catch(nlohmann::json::exception& e) {
+		std::cerr << "Failed to find root-level 'cell_id' element (must be an integer)." << std::endl;
+		return;
+		}
 	DataCell top(id, DataCell::CellType::document);
 	DTree::iterator doc_it = doc.set_head(top);
 
@@ -396,6 +455,10 @@ void cadabra::JSON_deserialise(const std::string& cj, DTree& doc)
 	
 	// Scan through json file.
 	const nlohmann::json& cells = root["cells"];
+	if(root.count("flags")>0) {
+		const nlohmann::json& flags = root["flags"];
+		doc.hide_input_cells = flags.value("hide_input_cells", false);
+		}
 	JSON_in_recurse(doc, doc_it, cells);
 	}
 
@@ -485,6 +548,14 @@ void cadabra::JSON_in_recurse(DTree& doc, DTree::iterator loc, const nlohmann::j
 				DataCell dc(id, cadabra::DataCell::CellType::image_png, textbuf.get<std::string>(), hide);
 				last=doc.append_child(loc, dc);
 				}
+			else if(cell_type=="image_svg") {
+				DataCell dc(id, cadabra::DataCell::CellType::image_svg, textbuf.get<std::string>(), hide);
+				last=doc.append_child(loc, dc);
+				}
+			else if(cell_type=="slider") {
+				DataCell dc(id, cadabra::DataCell::CellType::slider, textbuf.get<std::string>(), hide);
+				last=doc.append_child(loc, dc);
+				}
 			else {
 				std::cerr << "cadabra-client: found unknown cell type '"+cell_type+"', ignoring" << std::endl;
 				continue;
@@ -558,15 +629,19 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		case DataCell::CellType::latex_view:
 			break;
 		case DataCell::CellType::error:
+			str << "{\\color{red}";
+			break;
+		case DataCell::CellType::slider:
 			break;
 		case DataCell::CellType::input_form:
 			break;
-		case DataCell::CellType::image_png:
+ 		case DataCell::CellType::image_png:
+ 		case DataCell::CellType::image_svg:
 			std::size_t pos=image_file_base.rfind('/');
 			std::string fileonly=image_file_base.substr(pos+1);
-			str << "\\begin{center}\n\\includegraphics[width=.6\\textwidth]{"
+			str << "\\begin{minipage}{\\textwidth}\n\\includegraphics[width=.6\\textwidth]{"
 			    << fileonly+std::to_string(image_num)+"}\n"
-			    << "\\end{center}\n";
+			    << "\\end{minipage}\n";
 			break;
 		}
 
@@ -577,32 +652,52 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		out << base64_decode(it->textbuf);
 		++image_num;
 		}
+	else if(it->cell_type==DataCell::CellType::image_svg) {
+		// Images have to be saved to disk as separate files as
+		// LaTeX has no concept of images embedded in the .tex file.
+		std::string basename=image_file_base+std::to_string(image_num);
+		std::ofstream out(basename+".svg");
+		out << base64_decode(it->textbuf);
+		out.close();
+		// Convert to pdf, otherwise LaTeX cannot include it.
+      //		inkscape t.svg --export-pdf=t.pdf
+		int res = system((std::string("inkscape ")+basename+std::string(".svg --export-type=pdf --export-filename=")+basename+std::string(".pdf")).c_str());
+		if(res!=0)
+			throw std::logic_error("DataCell::LaTeX_recurse: failed to run 'convert' to convert svg to pdf.");
+		++image_num;
+		}
 	else {
 		if(it->textbuf.size()>0) {
 			if(it->cell_type!=DataCell::CellType::document
-			      && it->cell_type!=DataCell::CellType::latex
-			      && it->cell_type!=DataCell::CellType::input_form) {
-				std::string lr(it->textbuf);
-				// Make sure to sync these with the same in TeXEngine.cc !!!
-				lr=std::regex_replace(lr, std::regex(R"(\\left\()"),            "\\brwrap{(}{");
-				lr=std::regex_replace(lr, std::regex(R"(\\right\))"),           "}{)}");
-				lr=std::regex_replace(lr, std::regex(R"(\\left\[)"),            "\\brwrap{[}{");
-				lr=std::regex_replace(lr, std::regex(R"(\\left\.)"),            "\\brwrap{.}{");
-				lr=std::regex_replace(lr, std::regex(R"(\\right\])"),           "}{]}");
-				lr=std::regex_replace(lr, std::regex(R"(\\left\\\{)"),            "\\brwrap{\\{}{");
-				lr=std::regex_replace(lr, std::regex(R"(\\right\\\})"),           "}{\\}}");
-				lr=std::regex_replace(lr, std::regex(R"(\\right\.)"),            "}{.}");
+				&& it->cell_type!=DataCell::CellType::latex
+				&& it->cell_type!=DataCell::CellType::input_form) {
+
+				if(!doc.hide_input_cells ||
+					(it->cell_type!=DataCell::CellType::python && it->cell_type!=DataCell::CellType::latex)) {
+						
+					std::string lr(it->textbuf);
+					// Make sure to sync these with the same in TeXEngine.cc !!!
+					lr=std::regex_replace(lr, std::regex(R"(\\left\()"),            "\\brwrap{(}{");
+					lr=std::regex_replace(lr, std::regex(R"(\\right\))"),           "}{)}");
+					lr=std::regex_replace(lr, std::regex(R"(\\left\[)"),            "\\brwrap{[}{");
+					lr=std::regex_replace(lr, std::regex(R"(\\left\.)"),            "\\brwrap{.}{");
+					lr=std::regex_replace(lr, std::regex(R"(\\right\])"),           "}{]}");
+					lr=std::regex_replace(lr, std::regex(R"(\\left\\\{)"),            "\\brwrap{\\{}{");
+					lr=std::regex_replace(lr, std::regex(R"(\\right\\\})"),           "}{\\}}");
+					lr=std::regex_replace(lr, std::regex(R"(\\right\.)"),            "}{.}");
 //				lr=std::regex_replace(lr, std::regex(R"(\\begin\{verbatim\})"), "");
 //				lr=std::regex_replace(lr, std::regex(R"(\\end\{verbatim\})"),   "");
-				lr=std::regex_replace(lr, std::regex(R"(\\begin\{dmath\*\})"),  "\\begin{adjustwidth}{1em}{0cm}$");
-				lr=std::regex_replace(lr, std::regex(R"(\\end\{dmath\*\})"),    "$\\end{adjustwidth}");
-				lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");
-				lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");				
-				str << lr << "\n";
+					lr=std::regex_replace(lr, std::regex(R"(\\begin\{dmath\*\})"),  "\\begin{adjustwidth}{1em}{0cm}$");
+					lr=std::regex_replace(lr, std::regex(R"(\\end\{dmath\*\})"),    "$\\end{adjustwidth}");
+					lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");
+					lr=std::regex_replace(lr, std::regex(R"(\\algorithm\{(.*)_(.*)\})"), "\\algorithm{$1\\textunderscore{}$2}");
+					lr=std::regex_replace(lr, std::regex(R"(🍅)"), "*");
+					str << lr << "\n";
+					}
 				}
 			}
 		}
-
+	
 	switch(it->cell_type) {
 		case DataCell::CellType::python:
 		case DataCell::CellType::output:
@@ -612,12 +707,22 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 //			str << "\\end{verbatim}\n";
 			break;
 		case DataCell::CellType::document:
+			break;
 		case DataCell::CellType::latex:
+			break;
 		case DataCell::CellType::latex_view:
+			break;
 		case DataCell::CellType::input_form:
+			break;
 		case DataCell::CellType::error:
+			str << "}";
+			break;
 		case DataCell::CellType::image_png:
 			break;
+		case DataCell::CellType::image_svg:
+			break;
+		case DataCell::CellType::slider:
+			break;
 		}
 
 	if(doc.number_of_children(it)>0) {
@@ -642,6 +747,8 @@ void cadabra::LaTeX_recurse(const DTree& doc, DTree::iterator it, std::ostringst
 		case DataCell::CellType::input_form:
 		case DataCell::CellType::error:
 		case DataCell::CellType::image_png:
+		case DataCell::CellType::image_svg:
+		case DataCell::CellType::slider:
 			break;
 		}
 
@@ -658,7 +765,7 @@ std::string cadabra::export_as_python(const DTree& doc)
 void cadabra::python_recurse(const DTree& doc, DTree::iterator it, std::ostringstream& str)
 	{
 	if(it->cell_type==DataCell::CellType::document)
-		str << "#!/usr/local/bin/cadabra2\n";
+		str << "#!/usr/bin/env cadabra2\n";
 	else {
 		if(it->cell_type==DataCell::CellType::python) {
 			if(it->textbuf.size()>0) {
