@@ -99,6 +99,9 @@ DisplaySympy::DisplaySympy(const Kernel& kernel, const Ex& e)
 			{"\\dot",     "dot"},
 			{"\\ddot",    "ddot"},
 
+			{"\\Im",      "im"},
+			{"\\Re",      "re"},
+
 		// A few symbols are reserved by sympy.
 			{"N", "sympyN"},
 			{"O", "sympyO"},
@@ -239,21 +242,28 @@ void DisplaySympy::print_children(std::ostream& str, Ex::iterator it, int )
 void DisplaySympy::print_multiplier(std::ostream& str, Ex::iterator it)
 	{
 	bool suppress_star=false;
-	mpz_class denom=it->multiplier->get_den();
+	if(it->multiplier->is_rational()) {
+		mpq_class q = it->multiplier->get_rational();
+		mpz_class denom=q.get_den();
+		mpz_class numer=q.get_num();
 
-	if(denom!=1) {
-		if(false && it->multiplier->get_num()<0)
-			str << "(" << it->multiplier->get_num() << ")";
-		else
-			str << it->multiplier->get_num();
-		str << "/" << it->multiplier->get_den();
-		}
-	else if(*it->multiplier==-1) {
-		str << "-";
-		suppress_star=true;
+		if(denom!=1) {
+			if(false && numer<0)
+				str << "(" << numer << ")";
+			else
+				str << numer;
+			str << "/" << denom;
+			}
+		else if(*it->multiplier==-1) {
+			str << "-";
+			suppress_star=true;
+			}
+		else {
+			str << *it->multiplier;
+			}
 		}
 	else {
-		str << *it->multiplier;
+		str << it->multiplier->get_double();
 		}
 
 	if(!suppress_star && !(*it->name=="1"))
@@ -332,7 +342,14 @@ void DisplaySympy::print_parent_rel(std::ostream& str, str_node::parent_rel_t pr
 
 void DisplaySympy::dispatch(std::ostream& str, Ex::iterator it)
 	{
-	// The node names below should only be reserved node names; all others
+	if(std::holds_alternative<std::shared_ptr<NTensor>>(it->content)) {
+		throw ArgumentException("Cannot yet convert NTensor to SymPy expression.");
+		}
+	else if(std::holds_alternative<std::shared_ptr<NInterpolatingFunction>>(it->content)) {
+		throw ArgumentException("Cannot yet convert NInterpolatingFunction to SymPy expression.");
+		}
+	
+   // The node names below should only be reserved node names; all others
 	// should be looked up using properties. FIXME
 	if(*it->name=="\\prod")        print_productlike(str, it, "*");
 	else if(*it->name=="\\sum")    print_sumlike(str, it);
