@@ -18,13 +18,20 @@ You should have received a copy of the GNU General Public License
 
 */
 
+#include "Config.hh"
+#ifdef mimalloc_FOUND
+  #include "mimalloc-new-delete.h"
+#endif
+
 #include "Storage.hh"
 #include "Combinatorics.hh"
+#include "Compare.hh"
 #include "Exceptions.hh"
 #include <iomanip>
 #include <sstream>
 #include <locale>
 #include <codecvt>
+#include "NTensor.hh"
 
 #include <boost/functional/hash.hpp>
 
@@ -37,12 +44,18 @@ namespace cadabra {
 
 	long to_long(multiplier_t mul)
 		{
-		return mul.get_num().get_si();
+		if(!mul.is_rational())
+			throw ConsistencyException("Cannot convert double to long.");
+		
+		return mul.get_rational().get_num().get_si();
 		}
 
 	double to_double(multiplier_t mul)
 		{
-		return mul.get_d();
+		if(mul.is_rational())
+			return mul.get_rational().get_d();
+		else
+			return mul.get_double();
 		}
 
 	std::string to_string(long num)
@@ -75,6 +88,23 @@ namespace cadabra {
 		//	std::cout << "Ex copy constructor" << std::endl;
 		}
 
+//	Ex::Ex(Ex other)
+//		: std::enable_shared_from_this<Ex>(other), tree<str_node>(other), state_(result_t::l_no_action)
+//		{
+//		//	std::cout << "Ex copy constructor" << std::endl;
+//		}
+
+	Ex& Ex::operator=(Ex other)
+		{
+		std::swap(static_cast<tree<str_node>&>(*this),
+					 static_cast<tree<str_node>&>(other));
+		std::swap((*this).state_, other.state_);
+		std::swap((*this).history, other.history);
+		std::swap((*this).terms, other.terms);
+
+		return *this;
+		}
+
 	Ex::Ex(const std::string& str)
 		: state_(result_t::l_no_action)
 		{
@@ -88,6 +118,14 @@ namespace cadabra {
 		multiply(begin()->multiplier, val);
 		}
 
+	Ex::Ex(double val)
+		: state_(result_t::l_no_action)
+		{
+		std::ostringstream str;
+		str << val;
+		set_head(str_node(str.str()));
+		}
+
 	Ex::result_t Ex::state() const
 		{
 		return state_;
@@ -96,20 +134,28 @@ namespace cadabra {
 	void Ex::update_state(Ex::result_t newstate)
 		{
 		switch(newstate) {
-			case Ex::result_t::l_error:
-				state_=newstate;
-				break;
 			case Ex::result_t::l_applied:
+			case Ex::result_t::l_applied_no_new_dummies:
+			case Ex::result_t::l_cached:
+			case Ex::result_t::l_checkpointed:
+				// If we are in error state, preserve this.
 				if(state_!=Ex::result_t::l_error)
 					state_=newstate;
 				break;
-			default:
+			case Ex::result_t::l_error:
+				// Changing to error state is always possible,
+				// but it requires a reset_state to get out again.
+				state_=newstate;
+				break;
+			case Ex::result_t::l_no_action:
 				break;
 			}
 		}
 
 	void Ex::reset_state()
 		{
+		// Always use reset_state to clear error states; you cannot
+		// clear error states with update_state.
 		state_=Ex::result_t::l_checkpointed;
 		}
 
@@ -151,6 +197,25 @@ namespace cadabra {
 		return to_long(*(begin()->multiplier));
 		}
 
+	bool Ex::is_string() const
+		{
+		iterator it=begin();
+		if(it != end()) {
+			multiplier_t mult = *it->multiplier;
+			++it;
+			if(it==end() && mult==1)
+				return true;
+			}
+		return false;
+		}
+
+	bool Ex::equals(const std::string& str) const
+		{
+		if(is_string())
+			return (*begin()->name == str);
+		throw ConsistencyException("Expected string, found maths expression.");
+		}
+
 	bool Ex::is_empty() const
 		{
 		return begin()==end();
@@ -168,7 +233,8 @@ namespace cadabra {
 			//		else if(name[i]=='\\') res+="\\backslash{}";
 			else                   res+=name[i];
 			}
-		str << res;
+		if(res!="1")
+			str << res;
 
 		Ex::sibling_iterator beg=it.begin();
 		Ex::sibling_iterator fin=it.end();
@@ -306,7 +372,7 @@ namespace cadabra {
 
 	std::ostream& Ex::print_recursive_treeform(std::ostream& str, Ex::iterator it, unsigned int& num)
 		{
-		bool compact_tree=getenv("CDB_COMPACTTREE");
+		bool compact_tree=getenv("CADABRA_COMPACT_TREE");
 
 		Ex::sibling_iterator beg=it.begin();
 		Ex::sibling_iterator fin=it.end();
@@ -331,10 +397,16 @@ namespace cadabra {
 			else
 				str << "  " << *it->multiplier;
 			}
+		if(std::holds_alternative<std::shared_ptr<NTensor>>(it->content)) {
+			str << " " << *std::get<std::shared_ptr<NTensor>>(it->content);
+			}
+		else if(std::holds_alternative<std::shared_ptr<NInterpolatingFunction>>(it->content)) {
+			str << " (NInterpolatingFunction)";
+			}
 		//	str << "  (" << calc_hash(it) << ")";
 		//	str << "  (" << depth(it) << ")";
 		//	str << "  (" << it->fl.bracket << " " << &(*it) << ")";
-		str << "  (" << it->fl.bracket << " " << it.node << ")";
+//		str << "  (" << it->fl.bracket << " " << it.node << ")";
 		if(!compact_tree) str << std::endl;
 
 		while(beg!=fin) {
@@ -450,111 +522,111 @@ namespace cadabra {
 		return *nod->multiplier;
 		}
 
-	unsigned int Ex::equation_number(Ex::iterator it) const
-		{
-		iterator historynode=named_parent(it, "\\history");
-		unsigned int num=0;
-		iterator sit=begin();
-		//	long totravel=0;
-		while(sit!=end()) {
-			//		++totravel;
-			if(*sit->name=="\\history") {
-				++num;
-				if(historynode==sit) {
-					//				txtout << "had to travel " << totravel << std::endl;
-					return num;
-					}
-				}
-			sit.skip_children();
-			++sit;
-			}
-		return 0;
-		}
-
-	nset_t::iterator Ex::equation_label(Ex::iterator it) const
-		{
-		nset_t::iterator ret=name_set.end();
-
-		iterator sit=begin();
-		while(sit!=end()) {
-			if(*sit->name=="\\history") {
-				if(it==sit)
-					goto found;
-				iterator eit=begin(sit);
-				iterator endit=sit;
-				endit.skip_children();
-				++endit;
-				while(eit!=endit) {
-					if(it==eit)
-						goto found;
-					++eit;
-					}
-				sit.skip_children();
-				}
-			++sit;
-			}
-found:
-		if(sit!=end()) {
-			sibling_iterator lit=begin(sit);
-			while(lit!=end(sit)) {
-				if(*lit->name=="\\label") {
-					ret=begin(lit)->name;
-					break;
-					}
-				++lit;
-				}
-			}
-		return ret;
-		}
-
-	// Always returns the \\history node of the equation (i.e. the top node).
-	Ex::iterator Ex::equation_by_number(unsigned int i) const
-		{
-		iterator it=begin();
-		unsigned int num=1;
-		while(it!=end()) {
-			if(*it->name=="\\history") {
-				if(num==i) return it;
-				else       ++num;
-				}
-			it.skip_children();
-			++it;
-			}
-		return it;
-		//	if(num==number_of_siblings(begin()))
-		//		return end();
-		//	return it;
-		}
-
-	Ex::iterator Ex::equation_by_name(nset_t::iterator nit) const
-		{
-		unsigned int tmp;
-		return equation_by_name(nit, tmp);
-		}
-
-	Ex::iterator Ex::equation_by_name(nset_t::iterator nit, unsigned int& tmp) const
-		{
-		unsigned int num=0;
-		iterator it=begin();
-		while(it!=end()) {
-			if(*it->name=="\\history") {
-				++num;
-				sibling_iterator lit=begin(it);
-				while(lit!=end(it)) {
-					if(*lit->name=="\\label") {
-						if(begin(lit)->name==nit) {
-							tmp=num;
-							return it;
-							}
-						}
-					++lit;
-					}
-				}
-			it.skip_children();
-			++it;
-			}
-		return end();
-		}
+//	unsigned int Ex::equation_number(Ex::iterator it) const
+//		{
+//		iterator historynode=named_parent(it, "\\history");
+//		unsigned int num=0;
+//		iterator sit=begin();
+//		//	long totravel=0;
+//		while(sit!=end()) {
+//			//		++totravel;
+//			if(*sit->name=="\\history") {
+//				++num;
+//				if(historynode==sit) {
+//					//				txtout << "had to travel " << totravel << std::endl;
+//					return num;
+//					}
+//				}
+//			sit.skip_children();
+//			++sit;
+//			}
+//		return 0;
+//		}
+//
+//	nset_t::iterator Ex::equation_label(Ex::iterator it) const
+//		{
+//		nset_t::iterator ret=name_set.end();
+//
+//		iterator sit=begin();
+//		while(sit!=end()) {
+//			if(*sit->name=="\\history") {
+//				if(it==sit)
+//					goto found;
+//				iterator eit=begin(sit);
+//				iterator endit=sit;
+//				endit.skip_children();
+//				++endit;
+//				while(eit!=endit) {
+//					if(it==eit)
+//						goto found;
+//					++eit;
+//					}
+//				sit.skip_children();
+//				}
+//			++sit;
+//			}
+//found:
+//		if(sit!=end()) {
+//			sibling_iterator lit=begin(sit);
+//			while(lit!=end(sit)) {
+//				if(*lit->name=="\\label") {
+//					ret=begin(lit)->name;
+//					break;
+//					}
+//				++lit;
+//				}
+//			}
+//		return ret;
+//		}
+//
+//	// Always returns the \\history node of the equation (i.e. the top node).
+//	Ex::iterator Ex::equation_by_number(unsigned int i) const
+//		{
+//		iterator it=begin();
+//		unsigned int num=1;
+//		while(it!=end()) {
+//			if(*it->name=="\\history") {
+//				if(num==i) return it;
+//				else       ++num;
+//				}
+//			it.skip_children();
+//			++it;
+//			}
+//		return it;
+//		//	if(num==number_of_siblings(begin()))
+//		//		return end();
+//		//	return it;
+//		}
+//
+//	Ex::iterator Ex::equation_by_name(nset_t::iterator nit) const
+//		{
+//		unsigned int tmp;
+//		return equation_by_name(nit, tmp);
+//		}
+//
+//	Ex::iterator Ex::equation_by_name(nset_t::iterator nit, unsigned int& tmp) const
+//		{
+//		unsigned int num=0;
+//		iterator it=begin();
+//		while(it!=end()) {
+//			if(*it->name=="\\history") {
+//				++num;
+//				sibling_iterator lit=begin(it);
+//				while(lit!=end(it)) {
+//					if(*lit->name=="\\label") {
+//						if(begin(lit)->name==nit) {
+//							tmp=num;
+//							return it;
+//							}
+//						}
+//					++lit;
+//					}
+//				}
+//			it.skip_children();
+//			++it;
+//			}
+//		return end();
+//		}
 
 	bool Ex::is_hidden(iterator it) const
 		{
@@ -567,25 +639,25 @@ found:
 		return false;
 		}
 
-	Ex::iterator Ex::procedure_by_name(nset_t::iterator nit) const
-		{
-		iterator it=begin();
-		while(it!=end()) {
-			if(*it->name=="\\procedure") {
-				sibling_iterator lit=begin(it);
-				while(lit!=end(it)) {
-					if(*lit->name=="\\label") {
-						if(begin(lit)->name==nit)
-							return it;
-						}
-					++lit;
-					}
-				}
-			it.skip_children();
-			++it;
-			}
-		return end();
-		}
+//	Ex::iterator Ex::procedure_by_name(nset_t::iterator nit) const
+//		{
+//		iterator it=begin();
+//		while(it!=end()) {
+//			if(*it->name=="\\procedure") {
+//				sibling_iterator lit=begin(it);
+//				while(lit!=end(it)) {
+//					if(*lit->name=="\\label") {
+//						if(begin(lit)->name==nit)
+//							return it;
+//						}
+//					++lit;
+//					}
+//				}
+//			it.skip_children();
+//			++it;
+//			}
+//		return end();
+//		}
 
 	Ex::iterator Ex::replace_index(iterator pos, const iterator& from, bool keep_parent_rel)
 		{
@@ -642,59 +714,59 @@ found:
 		return pos;
 		}
 
-	unsigned int Ex::number_of_equations() const
-		{
-		unsigned int last_eq=0;
-		iterator eq=begin();
-		while(eq!=end()) {
-			if(*eq->name=="\\history")
-				++last_eq;
-			eq.skip_children();
-			++eq;
-			}
-		return last_eq;
-		}
+//	unsigned int Ex::number_of_equations() const
+//		{
+//		unsigned int last_eq=0;
+//		iterator eq=begin();
+//		while(eq!=end()) {
+//			if(*eq->name=="\\history")
+//				++last_eq;
+//			eq.skip_children();
+//			++eq;
+//			}
+//		return last_eq;
+//		}
 
-	Ex::iterator Ex::equation_by_number_or_name(iterator it, unsigned int last_used_equation,
-	      unsigned int& real_eqno) const
-		{
-		iterator ret;
-		if(it->is_rational()) {
-			int eqno=static_cast<int>(it->multiplier->get_d());
-			real_eqno=eqno;
-			ret=equation_by_number(eqno);
-			}
-		else {
-			if(*it->name=="%") {
-				ret=equation_by_number(last_used_equation);
-				real_eqno=last_used_equation;
-				}
-			else {
-				ret=equation_by_name(it->name, real_eqno);
-				}
-			}
-		return ret;
-		}
+//	Ex::iterator Ex::equation_by_number_or_name(iterator it, unsigned int last_used_equation,
+//	      unsigned int& real_eqno) const
+//		{
+//		iterator ret;
+//		if(it->is_rational()) {
+//			int eqno=static_cast<int>(it->multiplier->get_d());
+//			real_eqno=eqno;
+//			ret=equation_by_number(eqno);
+//			}
+//		else {
+//			if(*it->name=="%") {
+//				ret=equation_by_number(last_used_equation);
+//				real_eqno=last_used_equation;
+//				}
+//			else {
+//				ret=equation_by_name(it->name, real_eqno);
+//				}
+//			}
+//		return ret;
+//		}
 
-	Ex::iterator Ex::equation_by_number_or_name(iterator it, unsigned int last_used_equation) const
-		{
-		unsigned int tmp;
-		return equation_by_number_or_name(it, last_used_equation, tmp);
-		}
-
-	std::string Ex::equation_number_or_name(iterator it, unsigned int last_used_equation) const
-		{
-		std::stringstream ss;
-		if(it->is_rational()) {
-			int eqno=static_cast<int>(it->multiplier->get_d());
-			ss << eqno;
-			}
-		else {
-			if(*it->name=="%") ss << last_used_equation;
-			else               ss << *it->name;
-			}
-		return ss.str();
-		}
+//	Ex::iterator Ex::equation_by_number_or_name(iterator it, unsigned int last_used_equation) const
+//		{
+//		unsigned int tmp;
+//		return equation_by_number_or_name(it, last_used_equation, tmp);
+//		}
+//
+//	std::string Ex::equation_number_or_name(iterator it, unsigned int last_used_equation) const
+//		{
+//		std::stringstream ss;
+//		if(it->is_rational()) {
+//			int eqno=static_cast<int>(it->multiplier->get_d());
+//			ss << eqno;
+//			}
+//		else {
+//			if(*it->name=="%") ss << last_used_equation;
+//			else               ss << *it->name;
+//			}
+//		return ss.str();
+//		}
 
 	bool Ex::operator==(const Ex& other) const
 		{
@@ -728,6 +800,7 @@ found:
 		}
 
 	str_node::str_node(void)
+		: content(std::monostate())
 		{
 		multiplier=rat_set.insert(1).first;
 		//	fl.modifier=m_none;
@@ -736,7 +809,24 @@ found:
 		//	fl.mark=0;
 		}
 
+//	str_node::str_node(const str_node& other)
+//		: name(other.name)
+//		, multiplier(other.multiplier)
+//		, content(other.content)
+//		, fl(other.fl)
+//		{
+//		}
+//	
+//	str_node::str_node(str_node&& other)
+//		: name(std::move(other.name))
+//		, multiplier(std::move(other.multiplier))
+//		, content(std::move(other.content))
+//		, fl(std::move(other.fl))
+//		{
+//		}
+	
 	str_node::str_node(nset_t::iterator nm, bracket_t br, parent_rel_t pr)
+		: content(std::monostate())
 		{
 		multiplier=rat_set.insert(1).first;
 		name=nm;
@@ -747,6 +837,7 @@ found:
 		}
 
 	str_node::str_node(const std::u32string& nm, bracket_t br, parent_rel_t pr)
+		: content(std::monostate())
 		{
 #ifdef _MSC_VER
 		std::wstring_convert<std::codecvt_utf8<int32_t>, int32_t> conv;
@@ -768,6 +859,7 @@ found:
 		}
 
 	str_node::str_node(const std::string& nm, bracket_t br, parent_rel_t pr)
+		: content(std::monostate())
 		{
 		multiplier=rat_set.insert(1).first;
 		name=name_set.insert(nm).first;
@@ -784,6 +876,11 @@ found:
 		else throw std::logic_error("flip_parent_rel called on non-index");
 		}
 
+	bool str_node::is_standard() const
+		{
+		return std::holds_alternative<std::monostate>(content);
+		}
+
 	bool str_node::is_zero() const
 		{
 		if(*multiplier==0) return true;
@@ -792,20 +889,28 @@ found:
 
 	bool str_node::is_identity() const
 		{
-		if(*name=="1" && *multiplier==1) return true;
+		if(*name=="1" && *multiplier==1 && is_standard()) return true;
 		return false;
 		}
 
 	bool str_node::is_rational() const
 		{
-		return (*name=="1");
+		return (*name=="1" && multiplier->is_rational() && is_standard());
+		}
+
+	bool str_node::is_double() const
+		{
+		return (*name=="1" && multiplier->is_double() && is_standard());
 		}
 
 	bool str_node::is_integer() const
 		{
 		if(*name=="1") {
-			if(multiplier->get_den()==1)
-				return true;
+			if(multiplier->is_rational() && is_standard()) {
+				auto r = multiplier->get_rational();
+				if(r.get_den()==1)
+					return true;
+				}
 			}
 		return false;
 		}
@@ -820,6 +925,18 @@ found:
 		return true;
 		}
 
+	bool str_node::is_unsimplified_double() const
+		{
+		if((*name).size()==0) return false;
+		bool founddot=false;
+		for(unsigned int i=0; i<(*name).size(); ++i) {
+			if(!isdigit((*name)[i]) && (*name)[i]!='.' && (*name)[i]!='e' && (*name)[i]!='-')
+				return false;
+			if((*name)[i]=='.') founddot=true;
+			}
+		return founddot;
+		}
+
 	bool str_node::is_unsimplified_integer() const
 		{
 		if((*name).size()==0) return false;
@@ -836,8 +953,6 @@ found:
 		return false;
 		}
 
-
-
 	bool str_node::is_quoted_string() const
 		{
 		if((*name).size()<2) return false;
@@ -1014,6 +1129,14 @@ found:
 		num=rat_set.insert(fac).first;
 		}
 
+	void divide(rset_t::iterator& num, multiplier_t fac)
+		{
+		multiplier_t tmp(*num);
+		tmp /= fac;
+		tmp.canonicalize();
+		num=rat_set.insert(tmp).first;
+		}
+
 	void set(rset_t::iterator& num, multiplier_t fac)
 		{
 		fac.canonicalize();
@@ -1039,7 +1162,7 @@ found:
 
 	void flip_sign(rset_t::iterator& num)
 		{
-		multiplier_t fac=-(*num);
+		multiplier_t fac = multiplier_t(-1)*(*num);
 		fac.canonicalize();
 		num=rat_set.insert(fac).first;
 		}
@@ -1054,7 +1177,8 @@ found:
 	bool str_node::operator<(const cadabra::str_node& other) const
 		{
 		if(*name<*other.name) return true;
-		else return false;
+		if(*name==*other.name && *multiplier<*other.multiplier) return true;
+		return false;
 		}
 
 	}
