@@ -9,10 +9,16 @@
 #define DBG_MACRO_DISABLE
 #include "dbg.h"
 
-// #define DEBUG 1
+// #define DEBUG "substitute.cc"
+#include "Debug.hh"
+
 
 using namespace cadabra;
 
+substitute::Rules substitute::replacement_rules = substitute::Rules();
+size_t substitute::cache_hits   = 0;
+size_t substitute::cache_misses = 0;
+
 substitute::substitute(const Kernel& k, Ex& tr, Ex& args_, bool partial)
 	: Algorithm(k, tr), comparator(k.properties), args(args_), sort_product_(k, tr), partial(partial)
 	{
@@ -21,68 +27,85 @@ substitute::substitute(const Kernel& k, Ex& tr, Ex& args_, bool partial)
 
 	// Stopwatch sw;
 	// sw.start();
-	cadabra::do_list(args, args.begin(), [&](Ex::iterator arrow) {
-		//args.print_recursive_treeform(std::cerr, arrow);
-		if(*arrow->name!="\\arrow" && *arrow->name!="\\equals")
-			throw ArgumentException("substitute: Argument is neither a replacement rule nor an equality.");
 
-		sibling_iterator lhs=args.begin(arrow);
-		sibling_iterator rhs=lhs;
-		rhs.skip_children();
-		++rhs;
+	// Check if args are present in global_rules
+	// bool skipchecks = k.replacement_rules->is_present(args);
+	bool skipchecks = replacement_rules.is_present(args);
+	if(skipchecks) ++cache_hits;
+	else           ++cache_misses;
 
-		if(*lhs->name=="") { // replacing a sub or superscript
-			lhs=tr.flatten_and_erase(lhs);
-			}
-		if(*rhs->name=="") { // replacing with a sub or superscript
-			rhs=tr.flatten_and_erase(rhs);
-			}
+	// std::cerr << cache_hits << " vs " << cache_misses << std::endl;
+	
+	// skip if args have already been checked
+	if (!skipchecks) {
+		cadabra::do_list(args, args.begin(), [&](Ex::iterator arrow) {
+			//args.print_recursive_treeform(std::cerr, arrow);
+			if(*arrow->name!="\\arrow" && *arrow->name!="\\equals")
+				throw ArgumentException("substitute: Argument is neither a replacement rule nor an equality.");
 
-		try {
-			if(*lhs->multiplier!=1) {
-				throw ArgumentException("substitute: No numerical pre-factors allowed on lhs of replacement rule.");
+			sibling_iterator lhs=args.begin(arrow);
+			sibling_iterator rhs=lhs;
+			rhs.skip_children();
+			++rhs;
+
+			if(*lhs->name=="") { // replacing a sub or superscript
+				lhs=tr.flatten_and_erase(lhs);
 				}
-			// test validity of lhs and rhs
-			iterator lhsit=lhs, stopit=lhs;
-			stopit.skip_children();
-			++stopit;
-			while(lhsit!=stopit) {
-				if(lhsit->is_object_wildcard()) {
-					if(tr.number_of_children(lhsit)>0) {
-						throw ArgumentException("substitute: Object wildcards cannot have child nodes.");
-						}
+			if(*rhs->name=="") { // replacing with a sub or superscript
+				rhs=tr.flatten_and_erase(rhs);
+				}
+
+			try {
+				if(*lhs->multiplier!=1) {
+					throw ArgumentException("substitute: No numerical pre-factors allowed on lhs of replacement rule.");
 					}
-				++lhsit;
-				}
-			lhsit=rhs;
-			stopit=rhs;
-			stopit.skip_children();
-			++stopit;
-			while(lhsit!=stopit) {
-				if(lhsit->is_object_wildcard()) {
-					if(tr.number_of_children(lhsit)>0) {
-						throw ArgumentException("substitute: Object wildcards cannot have child nodes.");
+				// test validity of lhs and rhs
+				iterator lhsit=lhs, stopit=lhs;
+				stopit.skip_children();
+				++stopit;
+				while(lhsit!=stopit) {
+					if(lhsit->is_object_wildcard()) {
+						if(tr.number_of_children(lhsit)>0) {
+							throw ArgumentException("substitute: Object wildcards cannot have child nodes.");
+							}
 						}
+					++lhsit;
+					}
+				lhsit=rhs;
+				stopit=rhs;
+				stopit.skip_children();
+				++stopit;
+				while(lhsit!=stopit) {
+					if(lhsit->is_object_wildcard()) {
+						if(tr.number_of_children(lhsit)>0) {
+							throw ArgumentException("substitute: Object wildcards cannot have child nodes.");
+							}
+						}
+					++lhsit;
 					}
-				++lhsit;
-				}
 
-			// check whether there are dummies.
-			index_map_t ind_free, ind_dummy;
-			classify_indices(lhs, ind_free, ind_dummy);
-			lhs_contains_dummies[arrow]= ind_dummy.size()>0;
-			ind_free.clear();
-			ind_dummy.clear();
-			if(rhs!=tr.end()) {
-				classify_indices(rhs, ind_free, ind_dummy);
-				rhs_contains_dummies[arrow]=ind_dummy.size()>0;
+				// check whether there are dummies.
+				index_map_t ind_free, ind_dummy;
+				classify_indices(lhs, ind_free, ind_dummy);
+				lhs_contains_dummies[arrow]= ind_dummy.size()>0;
+				ind_free.clear();
+				ind_dummy.clear();
+				if(rhs!=tr.end()) {
+					classify_indices(rhs, ind_free, ind_dummy);
+					rhs_contains_dummies[arrow]=ind_dummy.size()>0;
+					}
 				}
-			}
-		catch(std::exception& er) {
-			throw ArgumentException(std::string("substitute: Index error in replacement rule. ")+er.what());
-			}
-		return true;
-		});
+			catch(std::exception& er) {
+				throw ArgumentException(std::string("substitute: Index error in replacement rule. ")+er.what());
+				}
+			return true;
+			});
+		replacement_rules.store(args, lhs_contains_dummies, rhs_contains_dummies);
+		}
+	else {
+		replacement_rules.retrieve(args, lhs_contains_dummies, rhs_contains_dummies);
+		}
+
 	// sw.stop();
 	// std::cerr << "preparation took " << sw << std::endl;
 	}
@@ -120,20 +143,39 @@ bool substitute::can_apply(iterator st)
 
 		if(*lhs->name=="\\prod")     ret=comparator.match_subproduct(tr, lhs, tr.begin(lhs), st, conditions);
 		else if(*lhs->name=="\\sum") ret=comparator.match_subsum(tr, lhs, tr.begin(lhs), st, conditions);
-		else                         ret=comparator.match_subtree(tr, lhs, st, conditions);
+		else {
+			DEBUGLN( std::cerr << "substitute::can_apply: testing " << lhs << " against " << st << std::endl; );
+			ret=comparator.match_subtree(tr, lhs, st, conditions);
+			}
 
 		if(ret == Ex_comparator::match_t::subtree_match ||
-		      ret == Ex_comparator::match_t::match_index_less ||
-		      ret == Ex_comparator::match_t::match_index_greater) {
+			ret == Ex_comparator::match_t::match_index_less ||
+			ret == Ex_comparator::match_t::match_index_greater) {
 			use_rule=arrow;
 
 			// If we are not matching a partial sum or partial product, need to check that all
 			// terms or factors are accounted for.
 			if(!partial) {
-				dbg(comparator.factor_locations.size());
-				dbg(tr.number_of_children(st));
-				if(comparator.factor_locations.size()!=tr.number_of_children(st))
-					return args.end();
+				if(*lhs->name=="\\prod") {
+					if(*st->name!="\\prod")
+						return args.end();
+					
+					DEBUGLN(std::cerr << "substitute::can_apply: partial=false, so " 
+							  << comparator.factor_locations.size()
+							  << " has to equal "
+							  << tr.number_of_children(st) << std::endl;)
+					if(comparator.factor_locations.size()!=tr.number_of_children(st))
+						return args.end();
+					}
+				else {
+					// If lhs is not a product, then we need to check that the node in the
+					// expression we act on does not sit inside a product (and hence has
+					// more factors).
+					if(tr.is_head(st) || (*tr.parent(st)->name)!="\\prod")
+						return arrow;
+					else
+						return args.end();
+					}
 				}
 
 			return arrow;
@@ -151,9 +193,7 @@ bool substitute::can_apply(iterator st)
 
 Algorithm::result_t substitute::apply(iterator& st)
 	{
-#ifdef DEBUG
-	std::cerr << "substitute::apply at " << Ex(st) << std::endl;
-#endif
+	DEBUGLN( std::cerr << "substitute::apply at " << Ex(st) << std::endl; )
 
 //	dbg(comparator.replacement_map);
 //	for(auto& rule: comparator.replacement_map)
@@ -181,9 +221,13 @@ Algorithm::result_t substitute::apply(iterator& st)
 		//std::cerr << "rhs does not contain dummies" << std::endl;
 		}
 
-	// Replace all patterns on the rhs of the rule with the objects they matched.
+	// Stage 1: Replace all patterns on the rhs of the rule with the
+	// objects they matched.
+	
 	// Keep track of all indices which _have_ to stay what they are, in ind_forced.
 	// Keep track of insertion points of subtrees.
+	// NOTE: this does not yet insert the replacement into the main tree!
+
 	iterator it=repl.begin();
 	Ex_comparator::replacement_map_t::iterator loc;
 	Ex_comparator::subtree_replacement_map_t::iterator sloc;
@@ -205,9 +249,7 @@ Algorithm::result_t substitute::apply(iterator& st)
 		//std::cerr << "consider element of repl " << Ex(it) << std::endl;
 
 		if(loc!=comparator.replacement_map.end()) { // name wildcards
-#ifdef DEBUG
-			std::cerr << "wildcard replaced: " << loc->first << " -> " << loc->second << std::endl;
-#endif
+			DEBUGLN( std::cerr << "wildcard replaced: " << loc->first << " -> " << loc->second << std::endl; )
 
 			// When a replacement is made here, and the index is actually
 			// a dummy in the replacement, we screw up the ind_dummy
@@ -223,16 +265,21 @@ Algorithm::result_t substitute::apply(iterator& st)
 			if(is_stripped || (it->is_name_wildcard() && !it->is_index()) ) {
 				// a?_{i j k} type patterns should only replace the head
 				// TODO: should we replace brackets here too?
+				DEBUGLN( std::cerr << "stripped replacing " << it
+							<< " with " << (*loc).second.begin() << " * " << *it->multiplier << std::endl; );
 				it->name=(*loc).second.begin()->name;
+				// See the comment below about multipliers.
 				multiply(it->multiplier, *(*loc).second.begin()->multiplier);
 				it->fl=(*loc).second.begin()->fl;
-				// std::cerr << "replaced: \n" << it << std::endl;
 				}
 			else {
-				// Careful with the multiplier: the object has been matched to the pattern
-				// without taking into account the top-level multiplier. So keep the multiplier
-				// of the thing we are replacing.
+				// Consider "A? -> 3 A?". If "A?" matches "2 C", then the replacement
+				// map will contain "A? mapsto 2 C". The rhs of the rule contains
+				// "3 A?" which should be changed to "3*2 C". After that, this "6 C"
+				// will get inserted into the tree in Stage 2 below.
 				multiplier_t mt=*it->multiplier;
+				DEBUGLN( std::cerr << "replacing " << it
+							<< " with " << (*loc).second.begin() << " * " << mt << std::endl; );
 				it=tr.replace_index(it, (*loc).second.begin()); //, true);
 				multiply(it->multiplier, mt);
 				}
@@ -247,16 +294,16 @@ Algorithm::result_t substitute::apply(iterator& st)
 			//std::cerr << "srule : " << Ex(it) << std::endl;
 			multiplier_t tmpmult=*it->multiplier; // remember target multiplier
 			iterator tmp= tr.insert_subtree(it, (*sloc).second);
-#ifdef DEBUG
-			std::cerr << "subtree replaced: " << repl << std::endl;
-#endif
+
+			DEBUGLN( std::cerr << "subtree replaced: " << repl << std::endl; );
+			
 			tmp->fl.bracket=it->fl.bracket;
 			tmp->fl.parent_rel=it->fl.parent_rel; // ok?
 			it=tr.erase(it);
 			multiply(tmp->multiplier, tmpmult);
-#ifdef DEBUG
-			std::cerr << "subtree replaced 2: " << repl << std::endl;
-#endif
+			
+			DEBUGLN( std::cerr << "subtree replaced 2: " << repl << std::endl; );
+
 			subtree_insertion_points.push_back(tmp);
 			index_map_t ind_subtree_free, ind_subtree_dummy;
 			// FIXME: as in the name wildcard case above, we only need these
@@ -276,9 +323,9 @@ Algorithm::result_t substitute::apply(iterator& st)
 	// not replaced here, but rather in the next step.
 	// std::cerr << ind_dummy.size() << std::endl;
 	if(ind_dummy.size()>0) {
-#ifdef DEBUG
-		std::cerr << "avoid dummy clashes" << std::endl;
-#endif
+
+		DEBUGLN( std::cerr << "avoid dummy clashes" << std::endl; );
+
 		index_map_t must_be_empty;
 		determine_intersection(ind_forced, ind_dummy, must_be_empty);
 		index_map_t::iterator indit=must_be_empty.begin();
@@ -307,15 +354,15 @@ Algorithm::result_t substitute::apply(iterator& st)
 	// After all replacements have been done, we need to cleanup the
 	// replacement tree.
 
-#ifdef DEBUG
-	std::cerr << repl << std::endl;
-#endif
+	DEBUGLN( std::cerr << "replacement before cleanup:\n" << repl << std::endl; );
 	
 	cleanup_dispatch_deep(kernel, repl);
 
-#ifdef DEBUG
-	std::cerr << "after cleanup:\n" << repl << std::endl;
-#endif
+	DEBUGLN( std::cerr << "after cleanup:\n" << repl << std::endl; );
+
+
+	// Stage 2: Now we can insert the right-hand side of the rule into
+	// the original tree.
 
 	repl.begin()->fl.bracket=st->fl.bracket;
 	bool rename_replacement_dummies_called=false;
@@ -350,7 +397,7 @@ Algorithm::result_t substitute::apply(iterator& st)
 		for(unsigned int i=1; i<comparator.factor_locations.size(); ++i)
 			tr.erase(comparator.factor_locations[i]);
 
-		multiply(repl.begin()->multiplier, 1/comparator.term_ratio);
+		divide(repl.begin()->multiplier, comparator.term_ratio);
 
 		// no need to keep repl
 		iterator newtr=tr.move_ontop(iterator(comparator.factor_locations[0]),repl.begin());
@@ -363,10 +410,15 @@ Algorithm::result_t substitute::apply(iterator& st)
 
 		}
 	else {
-#ifdef DEBUG
-		std::cerr << "move " << repl << " on top of " << st << std::endl;
-#endif
-		multiply(repl.begin()->multiplier, *st->multiplier);
+		DEBUGLN( std::cerr << "move " << repl << " on top of " << st << std::endl; );
+
+		if(!lhs->is_name_wildcard()
+			&& !lhs->is_range_wildcard()
+			&& !lhs->is_siblings_wildcard()
+			&& !lhs->is_autodeclare_wildcard()
+			&& !lhs->is_object_wildcard())
+			multiply(repl.begin()->multiplier, *st->multiplier);
+		
 		auto keep_parent_rel=st->fl.parent_rel;
 		st=tr.move_ontop(st, repl.begin()); // no need to keep the original repl tree
 		st->fl.parent_rel=keep_parent_rel;
@@ -392,9 +444,7 @@ Algorithm::result_t substitute::apply(iterator& st)
 	//	// FIXME: still needed?
 	//	cleanup_dispatch(kernel, tr, st);
 
-#ifdef DEBUG
-	std::cerr << tr << std::endl;
-#endif
+	DEBUGLN( std::cerr << tr << std::endl; );
 
 	dbg(tr.begin());
 	dbg(subtree_insertion_points.size());
@@ -413,3 +463,99 @@ Algorithm::result_t substitute::apply(iterator& st)
 	return result_t::l_applied;
 	}
 
+
+size_t substitute::cache_size()
+	{
+	return replacement_rules.size();
+	}
+
+void substitute::Rules::store(Ex& rules,
+								std::map<iterator, bool>& lhs_contains_dummies,
+								std::map<iterator, bool>& rhs_contains_dummies)
+	{
+	try {
+		// if number of stored rules has grown large, clean them up.
+		if (properties.size() >= cleanup_threshold) {
+			cleanup();
+			}
+		// If that didn't fix it, double the cleanup_threshold up to max_size
+		if (cleanup_threshold != max_size && properties.size() >= cleanup_threshold) {
+			if (cleanup_threshold * 2 < max_size) {
+				cleanup_threshold *= 2;
+				} 
+			else {
+				cleanup_threshold = max_size;
+				}
+			}
+
+		std::weak_ptr<Ex> rules_ptr = rules.shared_from_this();
+		// properties[rules_ptr] = { lhs_contains_dummies, rhs_contains_dummies };'
+		properties.insert(rules_ptr, { lhs_contains_dummies, rhs_contains_dummies });
+		// Mark this expression as cached; any change will remove that state
+		// and ensure that we do not use the cached expression later.
+		rules.update_state(result_t::l_cached);
+		}
+	catch(const std::bad_weak_ptr& error) {
+		return;
+		}
+	}
+
+void substitute::Rules::retrieve(Ex& rules,
+								std::map<iterator, bool>& lhs_contains_dummies,
+								std::map<iterator, bool>& rhs_contains_dummies) const
+	{
+	// Rules::present is assumed to have been called to check that the rules are valid
+	std::weak_ptr<Ex> rules_ptr = rules.shared_from_this();
+	lhs_contains_dummies = properties.at(rules_ptr).first;
+	rhs_contains_dummies = properties.at(rules_ptr).second;
+	}
+
+
+bool substitute::Rules::is_present(Ex& rules) const
+	{
+	// Look to see if the rules are present in the map
+	
+	try {
+		std::weak_ptr<Ex> rules_ptr = rules.shared_from_this();
+		auto rule_it = properties.find(rules_ptr);
+		bool rule_found = (rule_it != properties.end());
+		if (!rule_found) return false;
+		
+		// rules should have l_cached set
+		bool rule_unchanged = (rules.state() == result_t::l_cached);
+		
+		// If rule has been changed, erase it.
+		if (!rule_unchanged) {
+			properties.erase(rule_it);
+			return false;
+			} 
+		else {
+			return true;
+			}
+		}
+	catch(const std::bad_weak_ptr& error) {
+		return false;
+		}
+	}
+
+size_t substitute::Rules::size() const
+	{
+	return properties.size();
+	}
+
+void substitute::Rules::cleanup()
+	{
+	// Erase rules that are pointing to garbage-collected Ex expressions
+	// or rules that have possibly been changed.
+	for (auto it = properties.begin(); it != properties.end(); ) {
+		if (it->first.expired()) {
+			it = properties.erase(it);
+			} 
+		else if (it->first.lock()->state() != result_t::l_cached) {
+			it = properties.erase(it);
+			}
+		else {
+			++it;
+			}
+		}
+	}
