@@ -4,6 +4,10 @@
 
 using namespace cadabra;
 
+ExteriorDerivative::~ExteriorDerivative()
+	{
+	}
+
 std::string ExteriorDerivative::name() const
 	{
 	return "ExteriorDerivative";
@@ -12,13 +16,13 @@ std::string ExteriorDerivative::name() const
 Ex ExteriorDerivative::degree(const Properties& props, Ex::iterator it) const
 	{
 	auto sib=Ex::begin(it);
-	long deg=1;
+	int deg=1;
 	while(sib!=Ex::end(it)) {
 		const DifferentialFormBase *db = props.get<DifferentialFormBase>(sib);
 		if(db) {
 			auto thisdeg = db->degree(props, sib);
 			if(thisdeg.is_rational()==false) {
-				std::cerr << thisdeg << std::endl;
+				// std::cerr << thisdeg << std::endl;
 				throw NotYetImplemented("Cannot yet compute non-numerical form degrees.");
 				}
 			deg += to_long(thisdeg.to_rational());
