@@ -11,7 +11,7 @@ std::string Tableau::name() const
 bool Tableau::parse(Kernel&, keyval_t& keyvals)
 	{
 	keyval_t::const_iterator kv=keyvals.find("dimension");
-	if(kv!=keyvals.end()) dimension=to_long(*(kv->second->multiplier));
+	if(kv!=keyvals.end()) dimension=to_long(*(kv->second.begin()->multiplier));
 	else dimension=-1;
 	return true;
 	}
