@@ -18,6 +18,7 @@
 #include "Cleanup.hh"
 #include "Bridge.hh"
 #include "SympyCdb.hh"
+#include "NTensor.hh"
 
 // Includes for display routines
 #include <sstream>
@@ -222,25 +223,37 @@ namespace cadabra {
 
 	Ex_ptr fetch_from_python(const std::string& nm, pybind11::object scope)
 		{
-		// std::cerr << "fetch from python " << nm << std::endl;
-		if (!scope_has(scope, nm.c_str())) {
-			// std::cerr << "not present" << std::endl;
+		if( !scope_has(scope, nm) ) {
 			return 0;
 			}
 		auto obj = scope[nm.c_str()];
+
+		// Try 'Ex'
 		try {
 			return obj.cast<Ex_ptr>();
 			}
 		catch (const pybind11::cast_error& e) {
-			try {
-				auto exnode = obj.cast<ExNode>();
-				auto ret = std::make_shared<Ex>(exnode.it);
-				return ret;
-				}
-			catch (const pybind11::cast_error& e) {
-				std::cout << nm << " is not of type cadabra.Ex or cadabra.ExNode" << std::endl;
-				}
 			}
+
+		// Try 'ExNode'
+		try {
+			auto exnode = obj.cast<ExNode>();
+			auto ret = std::make_shared<Ex>(exnode.it);
+			return ret;
+			}
+		catch (const pybind11::cast_error& e) {
+			}
+
+		// Try float or int.
+		if(pybind11::isinstance<py::int_>(obj)) {
+			auto ret = std::make_shared<Ex>(pybind11::cast<int>(obj));
+			return ret;
+			}
+		if(pybind11::isinstance<py::float_>(obj)) {
+			auto ret = std::make_shared<Ex>(pybind11::cast<float>(obj));
+			return ret;
+			}
+			
 		return 0;
 		}
 
@@ -278,7 +291,6 @@ namespace cadabra {
 		return str.str();
 		}
 
-
 	pybind11::object Ex_as_sympy(Ex_ptr ex)
 		{
 		// Generate a string which can be parsed by Sympy.
@@ -294,6 +306,11 @@ namespace cadabra {
 		return ret;
 		}
 
+	pybind11::object ExNode_as_sympy(const ExNode& exnode)
+		{
+		return Ex_as_sympy(exnode.ex);
+		}
+
 	std::string Ex_as_sympy_string(Ex_ptr ex)
 		{
 		// Check to see if the expression is a scalar without dummy indices.
@@ -486,12 +503,15 @@ namespace cadabra {
 		{
 		if (ex->begin() == ex->end())
 			throw ArgumentException("Expression is empty, no head.");
-		pybind11::object mpq = pybind11::module::import("gmpy2").attr("mpq");
 		auto m = *ex->begin()->multiplier;
 		//	return mpq(2,3);
 
-		pybind11::object mult = mpq(m.get_num().get_si(), m.get_den().get_si());
-		return mult;
+		if(m.is_rational()) {
+			pybind11::object mpq = pybind11::module::import("gmpy2").attr("mpq");
+			pybind11::object mult = mpq(m.get_rational().get_num().get_si(), m.get_rational().get_den().get_si());
+			return mult;
+			}
+		else return pybind11::cast(m.get_double());
 		}
 
 	pybind11::list terms(Ex_ptr ex)
@@ -564,6 +584,12 @@ namespace cadabra {
 		return ptr;
 		}
 
+	Ex_ptr Ex_from_float(double num, bool)
+		{
+		auto ptr = std::make_shared<Ex>(num);
+		return ptr;
+		}
+
 	void call_post_process(Kernel& kernel, Ex_ptr ex)
 		{
 		// Find the 'post_process' function, and if found, turn off
@@ -628,9 +654,10 @@ namespace cadabra {
 			.value("no_match_less", Ex_comparator::match_t::no_match_less)
 			.value("no_match_greater", Ex_comparator::match_t::no_match_greater);
 
-		pybind11::class_<Ex, Ex_ptr >(m, "Ex")
+		pybind11::class_<Ex, Ex_ptr >(m, "Ex", py::buffer_protocol())
 		.def(py::init(&Ex_from_string), py::arg("input_form"), py::arg("make_ref") = true, py::arg("kernel") = nullptr)
 		.def(py::init(&Ex_from_int), py::arg("num"), py::arg("make_ref") = true)
+		.def(py::init(&Ex_from_float), py::arg("num"), py::arg("make_ref") = true)
 		.def("__str__", &Ex_as_str)
 		.def("_latex_", &Ex_as_latex)
 		.def("__repr__", &Ex_as_repr)
@@ -659,6 +686,34 @@ namespace cadabra {
 		.def("copy", [](const Ex& ex) { return std::make_shared<Ex>(ex); })
 		.def("changed", &Ex::changed_state)
 		.def("cleanup", &Ex_cleanup)
+		.def("__array__", [](Ex& ex) {
+			if(std::holds_alternative<std::shared_ptr<NTensor>>(ex.begin()->content)) {
+				std::shared_ptr<NTensor> ntp = std::get<std::shared_ptr<NTensor>>(ex.begin()->content);
+            py::object py_ntensor = py::cast(*ntp);
+
+				if(!py::isinstance<py::buffer>(py_ntensor)) {
+					throw std::runtime_error("NTensor does not implement buffer protocol");
+					}
+				
+				py::buffer  py_buffer  = py_ntensor.cast<py::buffer>();
+				py::module_ np = py::module_::import("numpy");
+            
+            // Create a numpy array view of the buffer (no copy)
+            py::object array = np.attr("asarray")(py_ntensor);
+				return array;
+ 				}
+			else {
+				throw ConsistencyException("Expression cannot be converted to numerical array.");
+				}
+			})
+//		.def_buffer([](Ex &ex) -> py::buffer_info {
+//			throw std::runtime_error("Help?");
+//			})
+//		.def("__floordiv__", [](const Ex& self, py::function func) {
+//			   Ex expr_copy(self);
+//            py::object result = func(expr_copy);
+//            return result.cast<Ex>();
+//			})
 		.def("__hash__", [](const Ex& ex) { return ex.calc_hash(ex.begin()); })
 		.def("__add__", static_cast<Ex_ptr(*)(const Ex_ptr, const ExNode)>(&Ex_add), py::is_operator{})
 		.def("__add__", static_cast<Ex_ptr(*)(const Ex_ptr, const Ex_ptr)>(&Ex_add), py::is_operator{})
@@ -670,7 +725,7 @@ namespace cadabra {
 			[](py::tuple t) { return Ex_from_string(t[0].cast<std::string>(), true, get_kernel_from_scope()); }
 		));
 
-		pybind11::class_<ExNode>(m, "ExNode", "Iterator over elements of an Ex mathematical expression.")
+		pybind11::class_<ExNode>(m, "ExNode", "Iterator over elements of an Ex mathematical expression.", py::buffer_protocol())
 			.def("__iter__", &ExNode::iter)
 			.def("__next__", &ExNode::next, pybind11::return_value_policy::reference_internal)
 			.def("input_form",  &ExNode::input_form)
@@ -680,6 +735,7 @@ namespace cadabra {
 			.def("__setitem__", &ExNode::setitem_iterator)
 			.def("_latex_", &ExNode::_latex_)
 			.def("__str__", &ExNode::__str__)
+			.def("_sympy_", &ExNode_as_sympy)
 			.def("terms", &ExNode::terms, "Return an ExNode iterator over all terms at the level of the current ExNode.")
 			.def("factors", &ExNode::factors, "Return an ExNode iterator over all factors at the level of the current ExNode.")
 			.def("own_indices", &ExNode::own_indices, "Return an ExNode iterator over all indices which are not inherited from child nodes.")
@@ -699,6 +755,9 @@ namespace cadabra {
 			.def_property("name", &ExNode::get_name, &ExNode::set_name, "Set the name property of the node pointed to by the ExNode.")
 			.def_property("parent_rel", &ExNode::get_parent_rel, &ExNode::set_parent_rel)
 			.def_property("multiplier", &ExNode::get_multiplier, &ExNode::set_multiplier)
+			.def_buffer([](ExNode &exnode) -> py::buffer_info {
+				throw std::range_error("Do not call this on ExNode");
+				})
 			.def("__add__", [](ExNode a, Ex_ptr b) {
 								 return a.add_ex(b);
 								 }, pybind11::is_operator{});
