@@ -990,6 +990,9 @@ inline int win32read(int *c) {
                     case 'E': /* ctrl+e, go to the end of the line */
                         *c = 5;
                         return 1;
+						 case '\\':
+							 *c = 28;
+							 return 1;
                 }
 
                 /* Other Ctrl+KEYs ignored */
@@ -1062,7 +1065,9 @@ inline int win32_write(int fd, const void *buffer, unsigned int count) {
 
 #define LINENOISE_DEFAULT_HISTORY_MAX_LEN 100
 #define LINENOISE_MAX_LINE 4096
+#ifndef _WIN32
 static const char *unsupported_term[] = {"dumb","cons25","emacs",NULL};
+#endif
 static CompletionCallback completionCallback;
 
 #ifndef _WIN32
@@ -1109,6 +1114,7 @@ enum KEY_ACTION {
     CTRL_T = 20,        /* Ctrl-t */
     CTRL_U = 21,        /* Ctrl+u */
     CTRL_W = 23,        /* Ctrl+w */
+	 CTRL_BACKSLASH = 28,/* Ctrl-backslash */
     ESC = 27,           /* Escape */
     BACKSPACE =  127    /* Backspace */
 };
@@ -2087,6 +2093,11 @@ inline void linenoiseEditDeletePrevWord(struct linenoiseState *l) {
     refreshLine(l);
 }
 
+void debug_key(int code) {
+    fprintf(stderr, "Key pressed: %d (0x%x)\n", code, code);
+    fflush(stderr);
+}
+
 /* This function is the core of the line editing capability of linenoise.
  * It expects 'fd' to be already in "raw mode" so that every key pressed
  * will be returned ASAP to read().
@@ -2148,6 +2159,7 @@ inline int linenoiseEdit(int stdin_fd, int stdout_fd, char *buf, int buflen, con
             if (c == 0) continue;
         }
 
+		  // debug_key(c);
         switch(c) {
         case ENTER:    /* enter */
             if (!history.empty()) history.pop_back();
@@ -2271,6 +2283,8 @@ inline int linenoiseEdit(int stdin_fd, int stdout_fd, char *buf, int buflen, con
         case CTRL_W: /* ctrl+w, delete previous word */
             linenoiseEditDeletePrevWord(&l);
             break;
+		  case CTRL_BACKSLASH:
+				  throw std::logic_error("SIGQUIT: program terminated.");
         }
     }
     return l.len;
